from base64 import b64encode
from getpass import getpass
import os
from typing import Optional, Union

from packaging import version
import requests
from requests import Session
from requests.adapters import HTTPAdapter, Retry
from requests.cookies import RequestsCookieJar

from mstrio import config
from mstrio.api import authentication, exceptions, hooks, misc, projects
from mstrio.utils.helper import deprecation_warning
import mstrio.utils.helper as helper


def get_connection(workstation_data: dict, project_name: str = None, project_id: str = None,
                   application_name: str = None,
                   application_id: str = None) -> Union["Connection", None]:
    """Connect to environment without providing user's credentials.

    It is possible to provide `project_id` or `project_name` to select
    project. When both `project_id` and `project_name` are `None`,
    project selection is cleared. When both `project_id` and
    `project_name` are provided, `project_name` is ignored.
    Project can be also selected later by calling method
    `select_project` on Connection object.
    Args:
        workstation_data (object): object which is stored in a 'workstationData'
            variable within Workstation
        project_name (str, optional): name of project (aka project)
            to select
        project_id (str, optional): id of project (aka project)
            to select
        application_name(str, optional): deprecated. Use project_name instead.
        application_id(str, optional): deprecated. Use project_id instead.
    Returns:
        connection to I-Server or None is case of some error
    """
    if (application_name or application_id):
        helper.deprecation_warning(
            '`application_name` and `application_id`',
            '`project_name` and `project_id`',
            '11.3.4.101',  # NOSONAR
            False)
        project_name = project_name or application_name
        project_id = project_id or application_id

    try:
        print('Creating connection from Workstation Data object...', flush=True)
        # get base url from Workstation Data object
        base_url = workstation_data['defaultEnvironment']['url']
        # get headers from Workstation Data object
        headers = workstation_data['defaultEnvironment']['headers']
        # get cookies from Workstation Data object
        cookies = workstation_data['defaultEnvironment']['cookies']

        # prepare cookies for request
        jar = RequestsCookieJar()
        for c in cookies:
            cookie_values = {'domain': '', 'path': '/'}
            cookie_values.update(c)
            jar.set(cookie_values['name'], cookie_values['value'], domain=cookie_values['domain'],
                    path=cookie_values['path'])
    except Exception as e:
        print('Some error occurred while preparing data to get identity token:')
        print(e)
        return None

    # get identity token
    r = requests.post(base_url + 'api/auth/identityToken', headers=headers, cookies=jar)
    if r.ok:
        # create connection to I-Server
        return Connection(base_url, identity_token=r.headers['X-MSTR-IdentityToken'],
                          project_id=project_id, project_name=project_name)
    else:
        print('HTTP %i - %s, Message %s' % (r.status_code, r.reason, r.text), flush=True)
        return None


class Connection():
    """Connect to and interact with the MicroStrategy environment.

    Creates a connection object which is used in subsequent requests and
    manages the user's connection with the MicroStrategy REST and Intelligence
    Servers.

    Examples:
        >>> from mstrio import connection
        >>>
        >>> # connect to the environment and chosen project
        >>> conn = connection.Connection(
        >>>     base_url="https://demo.microstrategy.com/MicroStrategyLibrary",
        >>>     username="username",
        >>>     password="password",
        >>>     project_name="MicroStrategy Tutorial"
        >>> )
        >>> # disconnect
        >>> conn.close()

    Attributes:
        base_url: URL of the MicroStrategy REST API server.
        username: Username.
        project_name: Name of the connected MicroStrategy Applicaiton.
        project_id: Id of the connected MicroStrategy Applicaiton.
        login_mode: Authentication mode. Standard = 1 (default) or LDAP = 16.
        ssl_verify: If True (default), verifies the server's SSL certificates
            with each request.
        user_id: Id of the authenticated user
        user_full_name: Full name of the authenticated user
        user_initials: Initials of the authenticated user
        iserver_version: Version of the I-Server
        web_version: Version of the Web Server
    """

    def __init__(self, base_url: str, username: Optional[str] = None,
                 password: Optional[str] = None, project_name: Optional[str] = None,
                 project_id: Optional[str] = None, login_mode: int = 1, ssl_verify: bool = True,
                 certificate_path: Optional[str] = None, proxies: Optional[dict] = None,
                 identity_token: Optional[str] = None, verbose: bool = True,
                 application_name: Optional[str] = None, application_id: Optional[str] = None):
        """Establish a connection with MicroStrategy REST API.

        You can establish connection by either providing set of values
        (`username`, `password`, `login_mode`) or just `identity_token`.

        When both `project_id` and `project_name` are `None`,
        project selection is cleared. When both `project_id`
        and `project_name` are provided, `project_name` is ignored.

        Args:
            base_url (str): URL of the MicroStrategy REST API server.
                Typically of the form:
                "https://<mstr_env>.com/MicroStrategyLibrary/api"
            username (str, optional): Username
            password (str, optional): Password
            project_name (str, optional): Name of the project you intend
                to connect to (case-sensitive). Provide either Project ID
                or Project Name.
            project_id (str, optional): Id of the project you intend to
                connect to (case-sensitive). Provide either Project ID
                or Project Name.
            application_name (str, optional): deprecated. Use project_name
            instead.
            application_id (str, optional): deprecated. Use project_id instead.
            login_mode (int, optional): Specifies the authentication mode to
                use. Supported authentication modes are: Standard (1)
                (default) or LDAP (16)
            ssl_verify (bool, optional): If True (default), verifies the
                server's SSL certificates with each request
            certificate_path (str, optional): Path to SSL certificate file, if
                None and ssl_verify is True then the certificate will be looked
                for in current working directory
            proxies (dict, optional): Dictionary mapping protocol or protocol
                and host to the URL of the proxy (e.g. {'http': 'foo.bar:3128',
                'http://host.name': 'foo.bar:4012'})
            identity_token (str, optional): Identity token for delegated
                session. Used for connection initialized by GUI.
            verbose (bool, optional): True by default. Controls the amount of
                feedback from the I-Server.
        """
        if (application_id or application_name):
            deprecation_warning(
                "`application_id` and `application_name`",
                "`project_id` or `project_name`",
                "11.3.4.101",  # NOSONAR
                False)

        # set the verbosity globally
        config.verbose = True if verbose and config.verbose else False
        self.base_url = helper.url_check(base_url)
        self.username = username
        self.login_mode = login_mode
        self.certificate_path = certificate_path
        self.identity_token = identity_token
        self.session = self.__configure_session(ssl_verify, certificate_path, proxies)
        self._web_version = None
        self._iserver_version = None
        self._user_id = None
        self._user_full_name = None
        self._user_initials = None
        self.__password = password
        project_id = project_id or application_id
        project_name = project_name or application_name

        if self.__check_version():
            # save the version of IServer in config file
            config.iserver_version = self.iserver_version
            # delegate identity token or connect and create new sesssion
            self.delegate() if self.identity_token else self.connect()
            self.select_project(project_id, project_name)
        else:
            print("""This version of mstrio is only supported on MicroStrategy 11.1.0400 or higher.
                     \rCurrent Intelligence Server version: {}
                     \rCurrent MicroStrategy Web version: {}
                     """.format(self.iserver_version, self.web_version))
            helper.exception_handler(msg="MicroStrategy Version not supported.",
                                     exception_type=exceptions.VersionException)

    def connect(self) -> None:
        """Authenticates the user and creates a new connection with the
        Intelligence Server.

        If an active connection is detected, the session is renewed.
        """
        response = authentication.session_renew(connection=self)
        if not response.ok:
            response = authentication.login(connection=self)
            self.session.headers['X-MSTR-AuthToken'] = response.headers['X-MSTR-AuthToken']
            if config.verbose:
                print("Connection to MicroStrategy Intelligence Server has been established.")
        else:
            if config.verbose:
                print("Connection to MicroStrategy Intelligence Server was renewed.")

    def delegate(self) -> None:
        """Delegates identity token to get authentication token and connect to
        MicroStrategy Intelligence Server."""
        response = authentication.delegate(self, self.identity_token, whitelist=[('ERR003', 401)])
        if response.ok:
            self.session.headers['X-MSTR-AuthToken'] = response.headers['X-MSTR-AuthToken']
            if config.verbose:
                print("Connection with MicroStrategy Intelligence Server has been delegated.")
        else:
            print("Could not share existing connection session, please input credentials:")
            self.__prompt_credentials()
            self.connect()

    def get_identity_token(self) -> str:
        """Create new identity token using existing authentication token."""
        response = authentication.identity_token(self)
        return response.headers['X-MSTR-IdentityToken']

    def validate_identity_token(self) -> bool:
        """Validate the identity token."""
        validate = authentication.validate_identity_token(self, self.identity_token)
        return validate.ok

    def close(self) -> None:
        """Closes a connection with MicroStrategy REST API."""
        authentication.logout(connection=self)
        self.session.close()

        if config.verbose:
            print("Connection to MicroStrategy Intelligence Server has been closed")

    def renew(self) -> None:
        """Checks if the session is still alive.

        If so, renews the session and extends session expiration.
        """
        status = authentication.session_renew(connection=self)

        if status.status_code == 204:
            if config.verbose:
                print("Your connection to MicroStrategy Intelligence Server was renewed.")
        else:
            response = authentication.login(connection=self)
            self.session.headers['X-MSTR-AuthToken'] = response.headers['X-MSTR-AuthToken']
            if config.verbose:
                print("""Connection with MicroStrategy Intelligence Server was not active.
                         \rNew connection has been established.""")

    def status(self) -> bool:
        """Checks if the session is still alive.

        Raises:
            HTTPError if I-Server behaves unexpectedly
        """
        status = authentication.session_status(connection=self)

        if status.status_code == 200:
            print("Connection to MicroStrategy Intelligence Server is active.")
            return True
        else:
            print("Connection to MicroStrategy Intelligence Server is not active.")
            return False

    def select_project(self, project_id: Optional[str] = None,
                       project_name: Optional[str] = None) -> None:
        """Select project for the given connection based on project_id or
        project_name.

        When both `project_id` and `project_name` are `None`, project selection
        is cleared. When both `project_id` and `project_name` are provided,
        `project_name` is ignored.

        Args:
            project_id: id of project to select
            project_name: name of project to select

        Raises:
            ValueError: if project with given id or name does not exist
        """
        if project_id is None and project_name is None:
            self.project_id = None
            self.project_name = None
            if config.verbose:
                print("No project selected.")
            return None

        if project_id is not None and project_name is not None:
            tmp_msg = ("Both `project_id` and `project_name` arguments provided. "
                       "Selecting project based on `project_id`.")
            helper.exception_handler(msg=tmp_msg, exception_type=Warning)

        _projects = projects.get_projects(connection=self).json()
        if project_id is not None:
            # Find which project name matches the project ID provided
            tmp_projects = helper.filter_list_of_dicts(_projects, id=project_id)
            if not tmp_projects:
                self.project_id, self.project_name = None, None
                tmp_msg = (f"Error connecting to project with id: {project_id}. "
                           "Project with given id does not exist or user has no access.")
                raise ValueError(tmp_msg)
        elif project_name is not None:
            # Find which project ID matches the project name provided
            tmp_projects = helper.filter_list_of_dicts(_projects, name=project_name)
            if not tmp_projects:
                self.project_id, self.project_name = None, None
                tmp_msg = (f"Error connecting to project with name: {project_name}. "
                           "Project with given name does not exist or user has no access.")
                raise ValueError(tmp_msg)

        self.project_id = tmp_projects[0]['id']
        self.project_name = tmp_projects[0]['name']
        self.session.headers['X-MSTR-ProjectID'] = self.project_id

    def select_application(self, application_id: Optional[str] = None,
                           application_name: Optional[str] = None) -> None:
        deprecation_warning(
            "`select_application` method",
            "`select_project` method",
            "11.3.4.101",  # NOSONAR
            False)
        self.select_project(project_id=application_id, project_name=application_name)

    def _get_authorization(self) -> str:
        self.__prompt_credentials()
        credentials = "{}:{}".format(self.username, self.__password).encode('utf-8')
        encoded_credential = b64encode(credentials)
        auth = "Basic " + str(encoded_credential, 'utf-8')
        return auth

    def _validate_project_selected(self) -> None:
        if self.project_id is None:
            raise AttributeError(
                "Project not selected. Select project using `select_project` method.")

    def __prompt_credentials(self) -> None:
        self.username = self.username if self.username is not None else input("Username: ")
        self.__password = self.__password if self.__password is not None else getpass("Password: ")
        self.login_mode = self.login_mode if self.login_mode else input(
            "Login mode (1 - Standard, 16 - LDAP): ")

    def __check_version(self) -> bool:
        """Checks version of I-Server and MicroStrategy Web and store these
        variables."""

        def get_server_status():
            json_response = misc.server_status(self).json()
            try:
                iserver_version = json_response["iServerVersion"][:9]
            except KeyError:
                raise exceptions.IServerException(
                    "I-Server is currently unavailable. Please contact your administrator.")
            web_version = json_response.get("webVersion")
            web_version = web_version[:9] if web_version else None
            return iserver_version, web_version

        self._iserver_version, self._web_version = get_server_status()
        return version.parse(self.iserver_version) >= version.parse("11.1.0400")

    def __configure_session(self, verify, certificate_path, proxies, existing_session=None,
                            retries=2, backoff_factor=0.3):
        """Creates a shared requests.Session() object with configuration from
        the initialization. Additional parameters change how the HTTPAdapter is
        configured.

        Args:
            existing_session (requests.Session() object, optional): Optional
                existing Session object to configure
            retries (int, optional): number of retries for certain type of
                requests (i.e. get)
            backoff_factor (float, optional):  backoff factor to apply between
                attempts after the second try (most errors are resolved
                immediately by a second try without a delay). urllib3 will sleep
                for: {backoff factor} * (2 ^ ({number of total retries} - 1))
                seconds. If the backoff_factor is 0.1, then sleep() will sleep
                for [0.0s, 0.2s, 0.4s, ...] between retries. It will never be
                longer than Retry. BACKOFF_MAX. By default, backoff is disabled
                (set to 0).
        """
        session = existing_session or Session()
        session.proxies = proxies if proxies is not None else {}
        session.verify = self._configure_ssl(verify, certificate_path)

        response_hooks = []
        if config.debug:
            response_hooks.append(hooks.print_url)
        if config.save_responses:
            response_hooks.append(hooks.save_response)
        if response_hooks:
            session.hooks['response'] = response_hooks

        status_forcelist = frozenset([413, 429, 500, 503])  # retry after status

        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
            raise_on_status=False,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)

        return session

    @staticmethod
    def _configure_ssl(ssl_verify, certificate_path):

        def get_certs_from_cwd():
            cert_extensions = ['crt', 'pem', 'p12']
            return [file for file in os.listdir('.') if file.split('.')[-1] in cert_extensions]

        def find_cert_in_cwd(ssl_verify):
            certs = get_certs_from_cwd()
            if certs and ssl_verify:
                return certs[0]
            else:
                return ssl_verify

        if certificate_path:
            return certificate_path
        return find_cert_in_cwd(ssl_verify)

    def __get_user_info(self) -> None:
        response = authentication.get_info_for_authenticated_user(connection=self).json()
        self._user_id = response.get("id")
        self._user_full_name = response.get("fullName")
        self._user_initials = response.get("initials")

    @property
    def user_id(self) -> str:
        if not self._user_id:
            self.__get_user_info()
        return self._user_id

    @property
    def user_full_name(self) -> str:
        if not self._user_full_name:
            self.__get_user_info()
        return self._user_full_name

    @property
    def user_initials(self) -> str:
        if not self._user_initials:
            self.__get_user_info()
        return self._user_initials

    @property
    def web_version(self) -> str:
        return self._web_version

    @property
    def iserver_version(self) -> str:
        return self._iserver_version



from datetime import datetime
from enum import Enum
from functools import wraps
import pytz


class DatetimeFormats(Enum):
    FULLDATETIME = '%Y-%m-%dT%H:%M:%S.%f%z'
    DATE = '%Y-%m-%d'
    YMDHMSmS = '%Y-%m-%dT%H:%M:%S.%f%z'
    YMDHMS = '%Y-%m-%dT%H:%M:%S%z'
    YMD = '%Y-%m-%d'


def str_to_datetime(date: str, format_str: str) -> datetime:
    """Change date format to datetime, based on `format_str` provided.
    If `date` is already a datetime object, return it. Make the date aware."""
    if date is None:
        return date
    try:
        if not isinstance(date, datetime):
            if format_str.find('%z') != -1:
                # Localization needed. Check if provided.
                if date[-1] == 'Z':
                    date = date[:-1]
                    format_str = format_str[:-2]
                elif date.find('+') == -1:
                    format_str = format_str[:-2]
            date = datetime.strptime(date, format_str)
        try:  # Localize to utc if not yet localized
            return pytz.utc.localize(date)
        except ValueError:  # Already localized
            return date
    except (TypeError, ValueError):
        print('Error')
        return None


def _return_with_miliseconds(func):
    """If the date has miliseconds, return only 3 decimal places. Older servers
    need this to work. New can parse both 3 and 6 decimal places."""

    @wraps(func)
    def inner(*args, **kwargs):
        res = func(*args, **kwargs)
        if isinstance(res, str) and res.find('.') != -1:
            plus_pos = res.find('+')
            res = f"{res[:res.find('.') + 4]}{res[plus_pos:] if plus_pos != -1 else ''}"
        return res

    return inner


@_return_with_miliseconds
def datetime_to_str(date: datetime, format_str: str) -> str:
    """Get date string from datetime, based on `format_str` provided.
    If `date` is already a string, return it. Make the date aware."""
    if isinstance(date, str):
        return date
    try:
        # We need the date to be aware, or the string won't be accepted by API
        try:
            return pytz.utc.localize(date).strftime(format_str)
        except ValueError:  # Already localized
            return date.strftime(format_str)
    except (TypeError, AttributeError):
        return None


def _get_only_datetimeformat_map(string_to_date_map: dict) -> dict:
    """Return all entries that are of `DATETIMEFORMAT` enum type."""
    return {
        key: value
        for (key, value) in string_to_date_map.items()
        if isinstance(value, DatetimeFormats)
    }


def _solve_prefix_and_convert_date(func, name: str, date: str, string_to_date_map: dict,
                                   only_datetimefomat: bool = True):
    if only_datetimefomat:
        string_to_date_map = _get_only_datetimeformat_map(string_to_date_map)
    if f'_{name}' in string_to_date_map:
        format = string_to_date_map[f'_{name}'].value if isinstance(
            string_to_date_map[f'_{name}'], DatetimeFormats) else string_to_date_map[f'_{name}']
        return func(date, format)
    elif name in string_to_date_map:
        format = string_to_date_map[name].value if isinstance(
            string_to_date_map[name], DatetimeFormats) else string_to_date_map[name]
        return func(date, format)
    return date


def map_str_to_datetime(name: str, date: str, string_to_date_map: dict,
                        only_datetimefomat: bool = True) -> datetime:
    """Change date format to datetime, based on `string_to_date_map`
    conversion dict. All occurances of `DATETIMEFORMAT` Enum in
    `string_to_date_map` are converted to coresponding string values.
    If name is not found in `string_to_date_map`, returns date without changes.
    """
    return _solve_prefix_and_convert_date(str_to_datetime, name, date, string_to_date_map,
                                          only_datetimefomat)


def map_datetime_to_str(name: str, date: datetime, string_to_date_map: dict,
                        only_datetimefomat: bool = True) -> str:
    """Change date format to string, based on `string_to_date_map`
    conversion dict. All occurances of `DATETIMEFORMAT` Enum in
    `string_to_date_map` are converted to coresponding string values.
    If name is not found in `string_to_date_map`, returns date without changes.
    """
    return _solve_prefix_and_convert_date(datetime_to_str, name, date, string_to_date_map,
                                          only_datetimefomat)


def bulk_str_to_datetime(source: dict, string_to_date_map: dict,
                         only_datetimefomat: bool = True) -> dict:
    """Change all dates from `source` found in `string_to_date_map`
    to datetime format. If parameter is not found in `string_to_date_map`,
    it is returned without changes."""
    for key, val in source.items():
        source[key] = map_str_to_datetime(key, val, string_to_date_map, only_datetimefomat)
    return source


def bulk_datetime_to_str(source: dict, string_to_date_map: dict,
                         only_datetimefomat: bool = True) -> dict:
    """Change all dates from `source` found in `string_to_date_map`
    to string format. If parameter is not found in `string_to_date_map`,
    it is returned without changes."""
    for key, val in source.items():
        source[key] = map_datetime_to_str(key, val, string_to_date_map, only_datetimefomat)
    return source



from typing import List, Optional, Union

from mstrio.connection import Connection

from ..schedule import Schedule
from .base_subscription import Subscription
from .content import Content
from .delivery import CacheType, Delivery, LibraryCacheTypes, ShortcutCacheFormat


class CacheUpdateSubscription(Subscription):
    """Class representation of MicroStrategy Cache Update Subscription
    object."""

    def __init__(self, connection: Connection, subscription_id: str = None, project_id: str = None,
                 project_name: str = None):
        """Initialize CacheUpdateSubscription object, populates it with
        I-Server data if subscription_id is passed.
        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is
        omitted.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`
            subscription_id: ID of the subscription to be initialized
            project_id: Project ID
            project_name: Project name
        """
        if subscription_id:
            super().__init__(connection, subscription_id, project_id, project_name)

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        allow_delivery_changes: Optional[bool] = None,
        allow_personalization_changes: Optional[bool] = None,
        allow_unsubscribe: Optional[bool] = None,
        send_now: Optional[bool] = None,
        owner_id: Optional[str] = None,
        schedules: Optional[Union[str, List[str], Schedule, List[Schedule]]] = None,
        contents: Optional[Content] = None,
        recipients: Optional[Union[List[dict], List[str]]] = None,
        delivery: Optional[Union[Delivery, dict]] = None,
        delivery_expiration_date: Optional[str] = None,
        contact_security: bool = True,
        cache_cache_type: Union[CacheType, str] = CacheType.RESERVED,
        cache_shortcut_cache_format: Union[ShortcutCacheFormat,
                                           str] = ShortcutCacheFormat.RESERVED,
        cache_library_cache_types: List[Union[LibraryCacheTypes, str]] = [LibraryCacheTypes.WEB],
        cache_reuse_dataset_cache: bool = False,
        cache_is_all_library_users: bool = False,
        delivery_notification_enabled: bool = False,
        delivery_personal_notification_address_id: Optional[str] = None,
    ) -> "CacheUpdateSubscription":
        """Creates a new cache update subscription.

        Args:
            connection: a MicroStrategy connection object
            name: name of the subscription,
            project_id: project ID,
            project_name: project name,
            application_id: deprecated. Use project_id instead.
            application_name: deprecated. Use project_name instead.
            allow_delivery_changes: whether the recipients can change
                the delivery of the subscription,
            allow_personalization_changes: whether the recipients can
                personalize the subscription,
            allow_unsubscribe: whether the recipients can unsubscribe
                from the subscription,
            send_now: indicates whether to execute the subscription
                immediately,
            owner_id: ID of the subscription owner, by default logged in
                user ID,
            schedules:
                Schedules IDs or Schedule objects,
            contents: The content settings.
            recipients: list of recipients IDs
                or dicts,
            delivery: delivery object or dict
            delivery_expiration_date: expiration date of the subscription,
                format should be yyyy - MM - dd,
            contact_security: whether to use contact security for each
                contact group member
            cache_cache_type: [RESERVED, SHORTCUT, SHORTCUTWITHBOOKMARK]
            cache_shortcut_cache_format: [RESERVED, JSON, BINARY, BOTH]
            cache_library_cache_types: Set of library cache types,
                available types can be web, android, ios
            cache_reuse_dataset_cache: Whether to reuse dataset cache
            cache_is_all_library_users: Whether for all library users
            delivery_notification_enabled: Whether notification is enabled,
                notification applies to cache
            delivery_personal_notification_address_id: Notification details
        """
        return super()._Subscription__create(
            connection=connection,
            name=name,
            project_id=project_id,
            project_name=project_name,
            allow_delivery_changes=allow_delivery_changes,
            allow_personalization_changes=allow_personalization_changes,
            allow_unsubscribe=allow_unsubscribe,
            send_now=send_now,
            owner_id=owner_id,
            schedules=schedules,
            contents=contents,
            recipients=recipients,
            delivery=delivery,
            delivery_mode=Delivery.DeliveryMode.CACHE,
            delivery_expiration_date=delivery_expiration_date,
            contact_security=contact_security,
            cache_cache_type=cache_cache_type,
            cache_shortcut_cache_format=cache_shortcut_cache_format,
            cache_library_cache_types=cache_library_cache_types,
            cache_reuse_dataset_cache=cache_reuse_dataset_cache,
            cache_is_all_library_users=cache_is_all_library_users,
            delivery_notification_enabled=delivery_notification_enabled,
            delivery_personal_notification_address_id=delivery_personal_notification_address_id,
        )

    def alter(
        cls,
        name: Optional[str] = None,
        allow_delivery_changes: Optional[bool] = None,
        allow_personalization_changes: Optional[bool] = None,
        allow_unsubscribe: Optional[bool] = None,
        send_now: Optional[bool] = None,
        owner_id: Optional[str] = None,
        schedules: Union[str, List[str], Schedule, List[Schedule]] = None,
        contents: Content = None,
        recipients: Union[List[dict], List[str]] = None,
        delivery: Union[Delivery, dict] = None,
        custom_msg=None,
        delivery_expiration_date: Optional[str] = None,
        contact_security: bool = True,
        cache_cache_type: Union[CacheType, str] = CacheType.RESERVED,
        cache_shortcut_cache_format: Union[ShortcutCacheFormat,
                                           str] = ShortcutCacheFormat.RESERVED,
        cache_library_cache_types: List[Union[LibraryCacheTypes, str]] = [LibraryCacheTypes.WEB],
        cache_reuse_dataset_cache: bool = False,
        cache_is_all_library_users: bool = False,
        delivery_notification_enabled: bool = False,
        delivery_personal_notification_address_id: Optional[str] = None,
    ):
        """Alter the subscription.

        Args:
            name: name of the subscription,
            allow_delivery_changes: whether the recipients can change
                the delivery of the subscription,
            allow_personalization_changes: whether the recipients can
                personalize the subscription,
            allow_unsubscribe: whether the recipients can unsubscribe
                from the subscription,
            send_now: indicates whether to execute the subscription
                immediately,
            owner_id: ID of the subscription owner, by default logged in
                user ID,
            schedules:
                Schedules IDs or Schedule objects,
            contents: The content settings.
            recipients: list of recipients IDs
                or dicts,
            delivery: delivery object or dict
            delivery_expiration_date: expiration date of the subscription,
                format should be yyyy - MM - dd,
            contact_security: whether to use contact security for each
                contact group member
            cache_cache_type: [RESERVED, SHORTCUT,
                SHORTCUTWITHBOOKMARK]
            cache_shortcut_cache_format: [RESERVED, JSON, BINARY, BOTH]
            cache_library_cache_types: Set of library cache types,
                available types can be web, android, ios
            cache_reuse_dataset_cache: Whether to reuse dataset cache
            cache_is_all_library_users: Whether for all library users
            delivery_notification_enabled: Whether notification is enabled,
                notification applies to cache
            delivery_personal_notification_address_id: Notification details
        """
        return super().alter(
            name=name, allow_delivery_changes=allow_delivery_changes,
            allow_personalization_changes=allow_personalization_changes,
            allow_unsubscribe=allow_unsubscribe, send_now=send_now, owner_id=owner_id,
            schedules=schedules, contents=contents, recipients=recipients, delivery=delivery,
            custom_msg=custom_msg, delivery_expiration_date=delivery_expiration_date,
            contact_security=contact_security, cache_cache_type=cache_cache_type,
            cache_shortcut_cache_format=cache_shortcut_cache_format,
            cache_library_cache_types=cache_library_cache_types,
            cache_reuse_dataset_cache=cache_reuse_dataset_cache,
            cache_is_all_library_users=cache_is_all_library_users,
            delivery_notification_enabled=delivery_notification_enabled,
            delivery_personal_notification_address_id=delivery_personal_notification_address_id)



from typing import Dict, List, Optional, TYPE_CHECKING, Union

from mstrio.access_and_security.security_filter import (LogicOperator, PredicateElementList,
                                                        PredicateFilter, PredicateForm,
                                                        PredicateJointElementList)
from mstrio.utils.helper import Any, Dictable

if TYPE_CHECKING:
    from mstrio.access_and_security.security_filter import PredicateBase
    from mstrio.connection import Connection


class Qualification(Dictable):
    """The security filter definition written as an expression tree over
    predicate nodes.

    We do not attempt to represent the entire filter using expression nodes.
    Instead we use predicate nodes as leaves of the main tree expression.

    Most predicate nodes use a simple data structure. Security filter support
    five types of predicates, including: `PredicateCustomExpression`,
    `PredicateJointElementList`, `PredicateElementList`, `PredicateForm`,
    `PredicateFilter`. These simple predicates correspond to the most common
    qualifications that are used in security filters. It is possible to combine
    predicates using class `LogicOperator`.

    It is invalid for a security filter expression to contain no nodes.

    Attributes:
        tree (dict): security filter definition written as an expression tree
            over predicate nodes
        text (string): human readable description of the expression. It is
            generated from the current specification of the expression. This
            string will appear similar to a parsable description of the
            expression in the current user's locale. It is intended to be used
            to allow a user to easily distinguish between expressions. But this
            string cannot actually be used as input to a parser session because
            it does not contain hidden information about ambiguities in the
            parse text. Since this representation is not able to fully describe
            an expression, there is no point in the client ever sending it to
            the service.
        tokens (array of dicts):  optional array, used if the expression is to
            be presented as a stream of tokens. When this representation is
            used by the service we would expect that the tokens would either
            completely parse the expression, or would describe an error in the
            parsing. When this representation is used by the client the tokens
            would either correspond to fresh text from the user (a `client`
            token) or to parts of the text that the user has not edited
            (a `parsed` token).
    """

    def __init__(self, tree: Union["LogicOperator", "PredicateBase"], text: Optional[str] = None,
                 tokens: Optional[List[dict]] = None):
        self.tree = tree
        self.text = text
        self.tokens = tokens

    @classmethod
    def from_dict(cls, source: Dict[str, Any], connection: Optional["Connection"] = None):
        tree = source.get("tree")
        tree_type = tree.get("type")
        if tree_type == "predicate_form_qualification":
            tree = PredicateForm.from_dict(tree, connection)
        elif tree_type == "predicate_joint_element_list":
            tree = PredicateJointElementList.from_dict(tree, connection)
        elif tree_type == "predicate_filter_qualification":
            tree = PredicateFilter.from_dict(tree, connection)
        elif tree_type == "predicate_element_list":
            tree = PredicateElementList.from_dict(tree, connection)
        elif tree_type == "operator":
            tree = LogicOperator.from_dict(tree, connection)
        else:
            tree = None
        new_source = source.copy()
        new_source["tree"] = tree
        return super().from_dict(new_source, connection)



"""This is the demo script to show how administrator can manage schedules
and events.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import Connection
from mstrio.distribution_services import (Event, list_events, list_schedules, list_subscriptions,
                                          Schedule, ScheduleEnums, ScheduleTime, Subscription)

base_url = "https://<>/MicroStrategyLibrary/api"
username = "some_username"
password = "some_password"
conn = Connection(base_url, username, password, login_mode=1)

# get list of all schedules
all_schedules = list_schedules(connection=conn)
# get all event based schedules
event_based_schedules = list_schedules(connection=conn,
                                       schedule_type=Schedule.ScheduleType.EVENT_BASED.value)
# get all expired Schedules
expired_schedules = list_schedules(connection=conn, expired=True)

# get all events
events = list_events(connection=conn)

# get single Schedule and list its properties
schedule = Schedule(connection=conn, id='AA11BB22CC33DD44EE55FF6677889900')
scheudle = Schedule(connection=conn, name='On Database Load')
schedule.list_properties()

# get single Event and list its properties
event = Event(connection=conn, name='Event1')
event.list_properties()

# disable/enable(pass new schedule expire time)
schedule.disable()
schedule.enable("2021-12-31")

# alter schedule properties
schedule.alter(description="New description")
# altering schedule type is also possible;
# changing event based schedule to time based (using ScheduleTime class):
schedule_time = ScheduleTime.from_details(
    recurrence_pattern=ScheduleEnums.RecurrencePattern.MONTHLY,
    monthly_pattern=ScheduleEnums.MonthlyPattern.DAYS_OF_MONTH,
    days_of_month=['1', '3', '10', '28'], repeat_interval=2,
    execution_pattern=ScheduleEnums.ExecutionPattern.ONCE, execution_time='02:00:00')
schedule.alter(time=schedule_time)

# create time based schedule
new_schedule = Schedule.create(
    connection=conn, name='Test Schedule', schedule_type=Schedule.ScheduleType.TIME_BASED,
    start_date='2021-06-24', recurrence_pattern=ScheduleEnums.RecurrencePattern.DAILY,
    repeat_interval=2, daily_pattern=ScheduleEnums.DailyPattern.DAILY,
    execution_pattern=ScheduleEnums.ExecutionPattern.ONCE, execution_time='02:00:00')
# create event based schedule
new_schedule = Schedule(connection=conn, name='Test Schedule',
                        schedule_type=Schedule.ScheduleType.EVENT_BASED, event=event,
                        start_date='2021-06-24', stop_date='2022-01-01')

# delete schedule
# (Note: deleting schedule will also delete related subscriptions!)
schedule.delete()

# good practice is to check if subscription is using the schedule,
# which we want to delete and if so alter it before deletion
other_schedule = Schedule(connection=conn, name="Some other schedule")
sub1 = Subscription(connection=conn, subscription_id='AA11BB22CC33DD44EE55FF6677889900',
                    project_id='00FF99EE88DD77CC66BB55AA44332211')
sub1.alter(schedules=other_schedule)

# get a list of subscriptions using schedule under the given project
all_subscriptions = list_subscriptions(conn, project_name='Sample Project')
subs_with_schedule = [sub for sub in all_subscriptions if sub.schedules[0].id == schedule.id]

# delete all expired schedules - user will be prompted to confirm every deletion
# (Note: deleting schedule will also delete related subscriptions!)
expired_schedules = list_schedules(connection=conn, expired=True)
for sch in expired_schedules:
    sch.delete()

# changing schedules for multiple subscriptions can be easily done using simple
# for loop (batch execution substitute)
sub2 = Subscription(connection=conn, subscription_id='XX11BB22CC33DD44EE55FF6677889900',
                    project_id='00FF99EE88DD77CC66BB55AA44332211')
list_of_subscriptions = [sub1, sub2]
for subscription in list_of_subscriptions:
    subscription.alter(schedules=schedule)



"""This is the demo script to show how to manage datasources.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import Connection
from mstrio.server import Project
from mstrio.users_and_groups import User
from mstrio.datasources import (DatasourceType, ExecutionMode, DatasourceConnection,
                                DatasourceInstance, DatasourceLogin, DatasourceMap,
                                list_datasource_connections, list_available_dbms,
                                list_datasource_instances, list_datasource_logins,
                                list_datasource_mappings)

base_url = "https://<>/MicroStrategyLibrary/api"
username = "some_username"
password = "some_password"
conn = Connection(base_url, username, password, project_name="MicroStrategy Tutorial",
                  login_mode=1)

# create a datasource login
login = DatasourceLogin.create(connection=conn, name='login_name', username='username',
                               password=password, description='Description of a login')

# update a datasource login
login.alter(name='changed', description='Changed description of a login')

# list properties of a datasource login
login.list_properties()

# delete a datasource login
login.delete(force=True)

# list all datasource logins
list_datasource_logins(connection=conn)

# create a datasource connection
ds_conn = DatasourceConnection.create(connection=conn, name='conn_name',
                                      description='Description a connection',
                                      execution_mode=ExecutionMode.ASYNCH_CONNECTION,
                                      datasource_login=login)

# update a datasource connection
ds_conn.alter(name='changed', description='Changed description of a connection')

# list properties of a datasource connection
ds_conn.list_properties()

# delete a datasource connection
ds_conn.delete(force=True)

# list all datasource connections
list_datasource_connections(connection=conn)

# get a dbms by name
dbms = list_available_dbms(connection=conn, name='PostgreSQL')

# create a datasource instance
ds_instance = DatasourceInstance.create(connection=conn, name='instance_name',
                                        description='Description of a datasource', dbms=dbms,
                                        datasource_connection=ds_conn,
                                        datasource_type=DatasourceType.RESERVED)

# update a datasource instance
ds_instance.alter(name='changed', description='Changed description of a datasource instance',
                  table_prefix='new_table_prefix')

# list properties of a datasource instance
ds_instance.list_properties()

# delete a datasource instance
ds_instance.delete(force=True)

# list all datasources by environment
list_datasource_instances(connection=conn)

# get project by name
project = Project(connection=conn, name="MicroStrategy Tutorial")

# list all datasources by project
list_datasource_instances(connection=conn, project=project)

# list all datasources by datasource connection
list_datasource_instances(connection=conn, datasource_connection={"id": ds_conn.id})

# get a user
user = User(connection=conn, username='mstr')
# create a datasource map
ds_map = DatasourceMap.create(connection=conn, project=project, user=user, ds_connection=ds_conn,
                              datasource=ds_instance, login=login)

# initialise a datasource map
example_ds_map_id = '1234567890A1234567890A1234567890'
ds_map = DatasourceMap(connection=conn, id=example_ds_map_id)

# list all datasource maps
list_datasource_mappings(connection=conn)

# delete a datasource map
ds_map.delete(force=True)



from copy import deepcopy
import csv
from enum import Enum
import inspect
from os.path import join as joinpath
from pprint import pprint
from sys import version_info
from typing import Any, Callable, Dict, List, Optional, Tuple, TYPE_CHECKING, TypeVar, Union

import dictdiffer
from pandas import DataFrame
import stringcase

from mstrio import config
from mstrio.api import objects
from mstrio.api.exceptions import VersionException
from mstrio.connection import Connection
from mstrio.types import ObjectTypes, ObjectSubTypes, ExtendedType  # noqa
from mstrio.utils import helper
from mstrio.utils.acl import ACE, ACLMixin, Rights
from mstrio.utils.dependence_mixin import DependenceMixin
from mstrio.utils.time_helper import bulk_str_to_datetime, DatetimeFormats, map_str_to_datetime

if TYPE_CHECKING:
    from mstrio.server import Project

T = TypeVar("T")


class EntityBase(helper.Dictable):
    """This class is for objects that do not have a specified MSTR type.

    Attributes:
        connection: A MicroStrategy connection object
        id: Object ID
        type: Object type set to None
        name: Object name
    """
    _OBJECT_TYPE: ObjectTypes = ObjectTypes.NONE  # MSTR object type defined in ObjectTypes
    _REST_ATTR_MAP: Dict[str, str] = {}
    _API_GETTERS: Dict[Union[str, tuple], Callable] = {}
    _FROM_DICT_MAP: Dict[str, Callable] = {
        'type': ObjectTypes
    }  # map attributes to Enums and Composites
    _AVAILABLE_ATTRIBUTES: Dict[str, type] = {}  # fetched on runtime from all Getters
    _API_PATCH: dict = {}
    _PATCH_PATH_TYPES: Dict[str, type] = {}  # used in update_properties method
    _API_DELETE: Callable = staticmethod(objects.delete_object)

    def __init__(self, connection: Connection, object_id: str, **kwargs) -> None:
        self._init_variables(connection=connection, id=object_id, **kwargs)
        if config.fetch_on_init and self._find_func("id") is not None:
            self.fetch("id")
        if config.verbose:
            print(self)

    def _init_variables(self, **kwargs):
        """Initialize variables given kwargs.

        Note: attributes not accepted by any implementation of this function
            in the inheritance chain will be disregarded.
        """
        kwargs = self._rest_to_python(kwargs)
        # create _AVAILABLE_ATTRIBUTES map
        self._AVAILABLE_ATTRIBUTES.update({key: type(val) for key, val in kwargs.items()})
        self._connection = kwargs.get("connection")
        self._id = kwargs.get("id")
        self._type = self._OBJECT_TYPE if 'type' in self._FROM_DICT_MAP else kwargs.get(
            "type", self._OBJECT_TYPE)
        self.name = kwargs.get("name")
        self._altered_properties = dict()

    def fetch(self, attr: Optional[str] = None) -> None:
        """Fetch the latest object state from the I-Server.

        Note:
            This method can overwrite local changes made to the object.

        Args:
            attr: Attribute name to be fetched.
        Raises:
            ValueError: if `attr` cannot be fetched.
        """
        functions = self._API_GETTERS  # by default fetch all endpoints

        if attr is not None:  # if attr is specified fetch endpoint matched to the attribute name
            function = self._find_func(attr)
            if not function:
                raise ValueError(f"The attribute `{attr}` cannot be fetched for this object")
            else:
                functions = {attr: func for attr, func in functions.items() if func == function}

        for key, func in functions.items():  # call respective API getters
            param_value_dict = auto_match_args_entity(func, self)

            response = func(**param_value_dict)
            if response.ok:
                response = response.json()
                if type(response) == dict:
                    object_dict = {
                        key if isinstance(key, str) and len(response) == 1 else k: v
                        for k, v in response.items()
                    }
                    self._set_object(**object_dict)
                elif type(response) == list:
                    self._set_object(**{key: response})
                # TODO: consider changing camel_to_snake logic to work with
                # list of keys

            # keep track of fetched attributes
            self._add_to_fetched(key)

    def _add_to_fetched(self, keys: Union[str, tuple]) -> None:
        if isinstance(keys, str):
            keys = [keys]
        for key in keys:
            key = key[1:] if key.startswith("_") else key
            self._fetched_attributes.add(key)

    @classmethod
    def _find_func(cls, attr: str) -> Optional[Callable]:
        """Try to find API endpoint in `cls._API_GETTERS` responsible for chosen
        attribute.

        Returns: Function or None if function not found
        """
        if not isinstance(attr, str):
            raise TypeError("`attr` parameter has to be of type str")

        for attributes, func in cls._API_GETTERS.items():
            if isinstance(attributes, str):
                if attr == attributes:
                    return func
            elif isinstance(attributes, tuple):
                if attr in attributes:
                    return func
            else:
                raise NotImplementedError
        return None

    @classmethod
    def _rest_to_python(cls, response: dict) -> dict:
        """Map REST API field names to Python API field names."""
        for rest_name, python_name in cls._REST_ATTR_MAP.items():
            if rest_name in response:
                old = response.pop(rest_name)
                if python_name:
                    response[python_name] = old
        return response

    @classmethod
    def _python_to_rest(cls, response: dict) -> dict:
        """Map Python API field names to REST API field names."""
        for rest_name, python_name in cls._REST_ATTR_MAP.items():
            if python_name and python_name in response:
                response[rest_name] = response.pop(python_name)
        return response

    def _set_object(self, **kwargs) -> None:
        """Set object attributes programmatically by providing keyword args.
        Support ENUMs and creating component objects."""

        object_info = helper.camel_to_snake(kwargs)
        object_info = self._rest_to_python(object_info)

        properties = set(
            (elem[0]
             for elem in inspect.getmembers(self.__class__, lambda x: isinstance(x, property))))

        for key, val in object_info.items():  # type: ignore
            # if self is a composite, create component instance
            if key in self._FROM_DICT_MAP:
                # determine which constructor will be used
                if isinstance(self._FROM_DICT_MAP[key], DatetimeFormats):
                    val = map_str_to_datetime(key, val, self._FROM_DICT_MAP)
                elif isinstance(self._FROM_DICT_MAP[key], type(Enum)):
                    val = self._FROM_DICT_MAP[key](val)
                elif isinstance(self._FROM_DICT_MAP[key], list) and val is not None:
                    if isinstance(self._FROM_DICT_MAP[key][0], type(Enum)):
                        val = [self._FROM_DICT_MAP[key][0](v) for v in val]
                    else:
                        val = [
                            self._FROM_DICT_MAP[key][0](source=v, connection=self.connection)
                            for v in val
                        ]
                else:
                    val = self._FROM_DICT_MAP[key](source=val, connection=self.connection)

            # create _AVAILABLE_ATTRIBUTES map
            self._AVAILABLE_ATTRIBUTES.update({key: type(val)})

            # check if attr is read-only and if yes return '_' version of it
            if key in properties:
                key = "_" + key

            setattr(self, key, val)

    def list_properties(self) -> dict:
        """List all properties of the object."""
        if hasattr(self, "_API_GETTERS"):  # fetch attributes not loaded on init
            attr = [
                attr for attr in self._API_GETTERS.keys() if isinstance(attr, str)  # type: ignore
            ]
            for a in attr:
                try:
                    getattr(self, a)
                except VersionException:
                    pass

        properties = inspect.getmembers(self.__class__, lambda x: isinstance(x, property))
        properties = {elem[0]: elem[1].fget(self) for elem in properties}
        attributes = {key: val for key, val in vars(self).items() if not key.startswith('_')}
        attributes = {**properties, **attributes}

        return {
            key: attributes[key] for key in sorted(attributes, key=helper.sort_object_properties)
        }

    def to_dataframe(self) -> DataFrame:
        """Return a `DataFrame` object containing object properties."""
        return DataFrame.from_dict(self.list_properties(), orient='index', columns=['value'])

    def print(self) -> None:
        """Pretty Print all properties of the object."""
        if version_info.major >= 3 and version_info.minor >= 8:
            pprint(self.list_properties(), sort_dicts=False)  # type: ignore
        else:
            pprint(self.list_properties())

    @classmethod
    def from_dict(cls: T, source: Dict[str, Any], connection: Connection,
                  to_snake_case: bool = True) -> T:
        """Instantiate an object from response without calling any additional
        getters."""
        obj = cls.__new__(cls)  # Does not call __init__
        object_source = helper.camel_to_snake(source) if to_snake_case else source
        obj._init_variables(connection=connection, **object_source)
        return obj

    def __str__(self):
        if self.__dict__.get("name"):
            return "{} object named: '{}' with ID: '{}'".format(self.__class__.__name__, self.name,
                                                                self.id)
        else:
            return "{} object with ID: '{}'".format(self.__class__.__name__, self.id)

    def __repr__(self):
        param_value_dict = auto_match_args_entity(self.__init__, self, exclude=['self'],
                                                  include_defaults=False)
        params_list = []
        for param, value in param_value_dict.items():
            if param == "connection" and isinstance(value, Connection):
                params_list.append("connection")
            else:
                params_list.append(f"{param}={repr(value)}")
        formatted_params = ", ".join(params_list)
        return f"{self.__class__.__name__}({formatted_params})"

    def __hash__(self):
        return hash((self.id, self._OBJECT_TYPE.value))

    def __eq__(self, other):
        """Equals operator to compare if an entity is equal to another object.

        We define 2 Entities as being equal if:
            1. the other object is an Entity and also shares the same id
            2. the other object is a string and is equal to either the name or
               id of the Entity
        This allows us to search for Entities in a collection by specifying its
        id, name, or another Entity itself. For example:
            if "user" in users: ...
            if "28ECA8BB11D5188EC000E9ABCA1B1A4F" in users: ...
        """
        if isinstance(other, str):
            return self.id == other or self.name == other
        elif isinstance(other, type(self)):
            return self.id == other.id
        elif isinstance(other, dict):
            # check all items in other dict equal entity object attributes
            return all(
                (getattr(self, k) == v if hasattr(self, k) else False for k, v in other.items()))
        else:
            return NotImplemented  # don't attempt to compare against unrelated types

    @classmethod
    def to_csv(cls: T, objects: Union[T, List[T]], name: str, path: Optional[str] = None,
               properties: Optional[List[str]] = None) -> None:
        """Export MSTR objects to a csv file.

        Optionally, save only the object properties specified in the properties
        parameter.

        Args:
            objects: List of objects of the same type that will be exported
            name: name of the csv file ending with '.csv'
            path: path to the directory where the file will be saved
            properties: list of object attribute names that should be included
                in the exported file
        """
        file = joinpath(path, name) if path else name
        list_of_objects = []
        if not name.endswith('.csv'):
            msg = ("The file extension is different than '.csv', please note that using a "
                   "different extension might disrupt opening the file correctly.")
            helper.exception_handler(msg, exception_type=Warning)
        if isinstance(objects, cls):
            properties_dict = objects.list_properties()
            if properties:
                list_of_objects.append(
                    {key: value for key, value in properties_dict.items() if key in properties})
            else:
                list_of_objects.append(properties_dict)
        elif isinstance(objects, list):
            if not properties:
                properties = objects[0].list_properties().keys()
            for obj in objects:
                if isinstance(obj, cls):
                    list_of_objects.append({
                        key: value
                        for key, value in obj.list_properties().items()
                        if key in properties
                    })
                else:
                    helper.exception_handler(
                        f"Object '{obj}' of type '{type(obj)}' is not supported.",
                        exception_type=Warning)
        else:
            raise TypeError((f"Objects should be of type {cls.__name__} or "
                             f"list of {cls.__name__}."))

        with open(file, 'w') as f:
            fieldnames = list_of_objects[0].keys()
            w = csv.DictWriter(f, fieldnames=fieldnames)
            w.writeheader()
            w.writerows(list_of_objects)
        if config.verbose:
            print(f"Object exported successfully to '{file}'")

    def is_modified(self, to_list: bool = False) -> Union[bool, list]:
        # TODO decide if needed or just deprecate
        """Compare the current object to the object on I-Server.

        Args:
            to_list: If True, return a list of tuples with object differences
        """
        temp = deepcopy(self)
        temp.fetch()
        differences = list(dictdiffer.diff(temp.__dict__, self.__dict__))
        if len(differences) == 0:
            if config.verbose:
                print("There are no differences between local and remote '{}' object.".format(
                    ObjectTypes(self.type).name))
            return differences if to_list else False
        else:
            return differences if to_list else True

    def update_properties(self) -> None:
        """Save compatible local changes of the object attributes to the
        I-Server.

        Raises:
            requests.HTTPError: if I-Server raises exception
        """
        changes = {k: v[1] for k, v in self._altered_properties.items()}
        self._alter_properties(**changes)
        self._altered_properties.clear()

    def _send_proper_patch_request(self, properties: dict, op: str = 'replace') -> List[bool]:
        """Internal method to update objects with the specified patch wrapper.
        Used for adding and removing objects from nested properties of an
        object like memberships.

        Args:
            properties: dictionary of required changes
            op: operation type, 'replace' by default

        Returns:
            List of successful or unsuccessful requests.
        """
        changed = []
        camel_properties = helper.snake_to_camel(properties)
        for attrs, (func, func_type) in self._API_PATCH.items():
            body = {}
            if func_type == 'partial_put':
                for name, value in camel_properties.items():
                    if stringcase.snakecase(name) in attrs:
                        value = self._maybe_unpack(name, value, camel_case=True)
                        body[name] = self._validate_type(name, value)
            elif func_type == 'put':  # Update using the generic update_object()
                for name, value in properties.items():
                    if name in attrs:
                        setattr(self, name, self._validate_type(name, value))
                body = self.to_dict()
            elif func_type == 'patch':
                body = {"operationList": []}
                for name, value in camel_properties.items():
                    if stringcase.snakecase(name) in attrs:
                        value = self._maybe_unpack(name, value, camel_case=True)
                        body['operationList'].append({
                            "op": op,
                            "path": "/{}".format(name),
                            "value": self._validate_type(name, value)
                        })
            else:
                msg = f"{func} function is not supported by `_send_proper_patch_request`"
                raise NotImplementedError(msg)

            if not body:
                continue
            # send patch request from the specified update wrapper
            param_value_dict = auto_match_args_entity(func, self, exclude=["body"])
            param_value_dict['body'] = body
            response = func(**param_value_dict)

            if response.ok:
                changed.append(True)
                response = response.json()
                if type(response) == dict:
                    self._set_object(**response)
            else:
                changed.append(False)
        return changed

    def _alter_properties(self, **properties) -> None:
        """Generic alter method that has to be implemented in child classes
        where arguments will be specified."""
        if not properties:
            if config.verbose:
                print(f"No changes specified for {type(self).__name__} '{self.name}'.")
            return None

        changed = self._send_proper_patch_request(properties)

        if config.verbose and all(changed):
            msg = (f"{type(self).__name__} '{self.name}' has been modified on the server. "
                   f"Your changes are saved locally.")
            print(msg)

    def _update_nested_properties(self, objects, path: str, op: str,
                                  existing_ids: Optional[List[str]] = None) -> Tuple[str, str]:
        """Internal method to update objects with the specified patch wrapper.
        Used for adding and removing objects from nested properties of an
        object like memberships.

        Returns:
            IDs of succeeded and failed operations by filtering by existing IDs.
        """
        from mstrio.access_and_security.privilege import Privilege

        # check whether existing_ids are supplied
        if existing_ids is None:
            existing_ids = [obj.get('id') for obj in getattr(self, path)]

        # create list of objects from strings/objects/lists
        objects_list = objects if isinstance(objects, list) else [objects]
        object_map = {obj.id: obj.name for obj in objects_list if isinstance(obj, Entity)}

        object_ids_list = [
            obj.id if isinstance(obj, (EntityBase, Privilege)) else str(obj)
            for obj in objects_list
        ]

        # check if objects can be manipulated by comparing to existing values
        if op == "add":
            filtered_object_ids = sorted(
                list(filter(lambda x: x not in existing_ids, object_ids_list)))
        elif op == "remove":
            filtered_object_ids = sorted(list(filter(lambda x: x in existing_ids,
                                                     object_ids_list)))
        if filtered_object_ids:
            properties = {path: filtered_object_ids}
            self._send_proper_patch_request(properties, op)

        failed = list(sorted(set(object_ids_list) - set(filtered_object_ids)))
        failed_formatted = [object_map.get(object_id, object_id) for object_id in failed]
        succeeded_formatted = [
            object_map.get(object_id, object_id) for object_id in filtered_object_ids
        ]
        return (succeeded_formatted, failed_formatted)

    def _validate_type(self, name: str, value: T) -> T:
        """Validates whether the attribute is set using correct type.

        Raises:
            TypeError if incorrect.
        """
        type_map = {**self._AVAILABLE_ATTRIBUTES, **self._PATCH_PATH_TYPES}
        value_type = type_map.get(name, 'Not Found')

        if value_type != 'Not Found' and type(value) != value_type:
            raise TypeError(f"'{name}' has incorrect type. Expected type: '{value_type}'")
        return value

    def __setattr__(self, name: str, value: Any) -> None:
        """Overloads the __setattr__ method to validate if this attribute can
        be set for current object and verify value data types."""

        def track_changes():
            already_tracked = name in self._altered_properties
            if already_tracked:
                current_val = self._altered_properties[name][0]
            elif name in self.__dict__:
                current_val = self.__dict__[name]
            else:
                current_val = None
            is_changed = not (type(current_val) is type(value) and current_val == value)
            if is_changed:
                self._altered_properties.update({name: (current_val, value)})
            elif already_tracked:
                del self._altered_properties[name]

        # Keep track of changed properties if value is already fetched
        if hasattr(self, "_fetched_attributes"):
            if name in self._PATCH_PATH_TYPES and name in self._fetched_attributes:
                track_changes()
            # if value not equal to None then treat as fetched
            if value is not None:
                self._add_to_fetched(name)
        super(EntityBase, self).__setattr__(name, value)

    def __getattribute__(self, name: str) -> Any:
        """Fetch attributes if not fetched."""
        val = super(EntityBase, self).__getattribute__(name)

        if name in ["_fetched_attributes", "_find_func"]:
            return val
        if not hasattr(self, "_fetched_attributes"):
            self._fetched_attributes = set()
        if hasattr(self, "_fetched_attributes") and hasattr(self, "_find_func"):
            _name = name[1:] if name.startswith("_") else name
            was_fetched = _name in self._fetched_attributes
            can_fetch = self._find_func(_name) is not None and "id" in self._fetched_attributes
            if can_fetch and not was_fetched:
                self.fetch(_name)  # fetch the relevant object data
            val = super(EntityBase, self).__getattribute__(name)

        return val

    def _set_dates(self, **kwargs):
        """Transform all date strings provided in kwargs to datetime object.
        Date format strings have to be provided in `_FROM_DICT_MAP`"""
        kwargs = bulk_str_to_datetime(kwargs, self._FROM_DICT_MAP)
        for key, val in self._FROM_DICT_MAP.items():
            if isinstance(val, DatetimeFormats):
                if kwargs.get(key):
                    setattr(self, key, kwargs.get(key))
                elif len(key) > 1 and key[0] == '_' and kwargs.get(key[1:]):
                    setattr(self, key, kwargs.get(key[1:]))

    # TODO add docstrings
    @property
    def connection(self) -> Connection:
        return self._connection

    @property
    def id(self) -> str:
        return self._id

    @property
    def type(self) -> ObjectTypes:
        return self._type


class Entity(EntityBase, ACLMixin, DependenceMixin):
    """Base class representation of the MSTR object.

    Provides methods to fetch, update, and view the object. To implement
    this base class all class attributes have to be provided.

    Attributes:
        connection: A MicroStrategy connection object
        id: Object ID
        name: Object name
        description: Object description
        abbreviation: Object abbreviation
        type: Object type
        subtype: Object subtype
        ext_type: Object extended type
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        version: Version ID
        owner: Owner ID and name
        icon_path: Object icon path
        view_media: View media settings
        ancestors: List of ancestor folders
        certified_info: Certification status, time of certificaton, and
            information about the certifier (currently only for document and
            report)
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """

    _API_GETTERS: Dict[Union[str, tuple], Callable] = {
        ('id', 'name', 'description', 'abbreviation', 'type', 'subtype', 'ext_type',
         'date_created', 'date_modified', 'version', 'owner', 'icon_path', 'view_media',
         'ancestors', 'certified_info', 'acg', 'acl', 'comments', 'project_id', 'hidden',
         'target_info'): objects.get_object_info
    }
    _API_PATCH: dict = {
        ('name', 'description', 'abbreviation'): (objects.update_object, 'partial_put')
    }
    _PATCH_PATH_TYPES = {"name": str, "description": str, "abbreviation": str}
    _FROM_DICT_MAP = {
        **EntityBase._FROM_DICT_MAP,
        'ext_type': ExtendedType,
        'date_created': DatetimeFormats.FULLDATETIME,
        'date_modified': DatetimeFormats.FULLDATETIME,
        'acl': [ACE.from_dict],
        'acg': Rights,
    }

    def _init_variables(self, **kwargs) -> None:
        """Initialize variables given kwargs."""
        from mstrio.users_and_groups.user import User
        from mstrio.utils.certified_info import CertifiedInfo
        super(Entity, self)._init_variables(**kwargs)
        self._date_created = map_str_to_datetime("date_created", kwargs.get("date_created"),
                                                 self._FROM_DICT_MAP)
        self._date_modified = map_str_to_datetime("date_modified", kwargs.get("date_modified"),
                                                  self._FROM_DICT_MAP)
        self.description = kwargs.get("description")
        self.abbreviation = kwargs.get("abbreviation")
        self._subtype = kwargs.get("subtype")
        self._ext_type = ExtendedType(kwargs["ext_type"]) if kwargs.get("ext_type") else None
        self._version = kwargs.get("version")
        self._owner = User.from_dict(
            kwargs.get("owner"),
            self.connection,
        ) if kwargs.get("owner") else None
        self._icon_path = kwargs.get("icon_path")
        self._view_media = kwargs.get("view_media")
        self._ancestors = kwargs.get("ancestors")
        self._certified_info = CertifiedInfo.from_dict(
            kwargs.get("certified_info"),
            self.connection) if kwargs.get("certified_info") else None
        self._hidden = kwargs.get("hidden")
        self._project_id = kwargs.get("project_id")
        self._comments = kwargs.get("comments")
        self._target_info = kwargs.get("target_info")
        self._acg = Rights(kwargs.get("acg")) if kwargs.get("acg") else None
        self._acl = ([ACE.from_dict(ac, self._connection) for ac in kwargs.get("acl")]
                     if kwargs.get("acl") else None)

    # TODO add docstrings to all properties
    @property
    def subtype(self):
        return self._subtype

    @property
    def ext_type(self):
        return self._ext_type

    @property
    def date_created(self):
        return self._date_created

    @property
    def date_modified(self):
        return self._date_modified

    @property
    def version(self):
        return self._version

    @property
    def owner(self):
        return self._owner

    @property
    def icon_path(self):
        return self._icon_path

    @property
    def view_media(self):
        return self._view_media

    @property
    def ancestors(self):
        return self._ancestors

    @property
    def certified_info(self):
        return self._certified_info

    @property
    def acg(self):
        return self._acg

    @property
    def acl(self):
        return self._acl

    @property
    def hidden(self):
        return self._hidden

    @property
    def project_id(self):
        return self._project_id

    @property
    def comments(self):
        return self._comments

    @property
    def target_info(self):
        return self._target_info


class CopyMixin:
    """CopyMixin class adds creating copies of objects functionality.

    Currently project objects are not supported. Must be mixedin with
    Entity or its subclasses.
    """

    def create_copy(self: Entity, name: Optional[str] = None, folder_id: Optional[str] = None,
                    project: Optional[Union["Project", str]] = None,
                    application: Optional[Union["Project", str]] = None) -> Any:
        """Create a copy of the object on the I-Server.

        Args:
            name: New name of the object. If None, a default name is generated,
                such as 'Old Name (1)'
            folder_id: ID of the destination folder. If None, the object is
                saved in the same folder as the source object.
            project: By default, the project selected when
                creating Connection object. Override `project` to specify
                project where the current object exists.
            application: deprecated. Use project instead.

        Returns:
                New python object holding the copied object.
        """
        if application:
            helper.deprecation_warning(
                '`application`',
                '`project`',
                '11.3.4.101',  # NOSONAR
                False)
            project = project or application
        if self._OBJECT_TYPE.value in [32]:
            raise NotImplementedError("Object cannot be copied yet.")
        # TODO if object uniqueness depends on project_id extract proj_id
        # TODO automatically
        response = objects.copy_object(self.connection, id=self.id, name=name, folder_id=folder_id,
                                       object_type=self._OBJECT_TYPE.value, project_id=project)
        return self.from_dict(source=response.json(), connection=self.connection)


class DeleteMixin:
    """DeleteMixin class adds deleting objects functionality.
    Must be mixedin with Entity or its subclasses.
    """

    _delete_confirm_msg: str = None
    _delete_success_msg: str = None

    def delete(self: Entity, force: bool = False) -> bool:
        """Delete object.

        Args:
            force: If True, then no additional prompt will be shown before
                deleting object.

        Returns:
            True on success. False otherwise.
        """
        object_name = self.__class__.__name__

        user_input = 'N'
        if not force:
            message = self._delete_confirm_msg or (
                f"Are you sure you want to delete {object_name} "
                f"'{self.name}' with ID: {self._id}? [Y/N]: ")
            user_input = input(message) or 'N'
        if force or user_input == 'Y':
            param_value_dict = auto_match_args_entity(self._API_DELETE, self)
            response = self._API_DELETE(**param_value_dict)
            if response.status_code == 204 and config.verbose:
                print(self._delete_success_msg
                      or f"Successfully deleted {object_name} with ID: {self._id}.")
            return response.ok
        else:
            return False


class CertifyMixin:
    """CertifyMixin class adds certifying and decertifying functionality.
    Must be mixedin with Entity or its subclasses.
    """

    def _toggle_certification(self: Entity, certify: bool = True, success_msg: str = None) -> bool:
        object_name = self.__class__.__name__
        expected_result = 'certified' if certify else 'decertified'

        self.fetch()
        if certify == self.certified_info.certified:
            print(f"The {object_name} with ID: '{self._id}' is already {expected_result}")
            return True

        response = objects.toggle_certification(connection=self._connection, id=self._id,
                                                object_type=self._OBJECT_TYPE.value,
                                                certify=certify)
        if response.ok and config.verbose:
            self._set_object(**response.json())
            print(success_msg
                  or f"The {object_name} with ID: '{self._id}' has been {expected_result}.")
        return response.ok

    def certify(self: Entity) -> bool:
        """Certify object.
        Args:
           success_msg: Custom message displayed on success.
        Returns:
                True on success, False otherwise."""
        return self._toggle_certification(certify=True)

    def decertify(self: Entity) -> bool:
        """Decertify object.
        Args:
           success_msg: Custom message displayed on success.
        Returns:
                True on success, False otherwise."""
        return self._toggle_certification(certify=False)


class VldbMixin:
    """VLDBMixin class adds vldb management for supporting objects.

    Objects currently supporting VLDB settings are dataset, document, dossier.
    Must be mixedin with Entity or its subclasses.
    """
    _parameter_error = "Please specify the project parameter."

    def list_vldb_settings(self: Entity, project: Optional[str] = None,
                           application: Optional[str] = None) -> list:
        """List VLDB settings."""
        if application:
            helper.deprecation_warning(
                '`application`',
                '`project`',
                '11.3.4.101',  # NOSONAR
                False)
            project = project or application
        connection = self.connection if hasattr(self, 'connection') else self._connection
        if not project and connection.session.headers.get('X-MSTR-ProjectID') is None:
            raise ValueError(self._parameter_error)

        response = objects.get_vldb_settings(connection, self.id, self._OBJECT_TYPE.value, project)
        return response.json()

    def alter_vldb_settings(self: Entity, property_set_name: str, name: str, value: dict,
                            project: Optional[str] = None,
                            application: Optional[str] = None) -> None:
        """Alter VLDB settings for a given property set."""
        if application:
            helper.deprecation_warning(
                '`application`',
                '`project`',
                '11.3.4.101',  # NOSONAR
                False)
            project = project or application

        connection = self.connection if hasattr(self, 'connection') else self._connection
        if not project and connection.session.headers.get('X-MSTR-ProjectID') is None:
            raise ValueError(self._parameter_error)

        body = [{"name": name, "value": value}]
        response = objects.set_vldb_settings(connection, self.id, self._OBJECT_TYPE.value,
                                             property_set_name, body, project)
        if config.verbose and response.ok:
            print("VLDB settings altered")

    def reset_vldb_settings(self: Entity, project: Optional[str] = None,
                            application: Optional[str] = None) -> None:
        """Reset VLDB settings to default values."""
        if application:
            helper.deprecation_warning(
                '`application`',
                '`project`',
                '11.3.4.101',  # NOSONAR
                False)
            project = project or application

        connection = self.connection if hasattr(self, 'connection') else self._connection
        if not project and connection.session.headers.get('X-MSTR-ProjectID') is None:
            raise ValueError(self._parameter_error)

        response = objects.delete_vldb_settings(connection, self.id, self._OBJECT_TYPE.value,
                                                project)
        if config.verbose and response.ok:
            print("VLDB settings reset to default")


def auto_match_args_entity(func: Callable, obj: EntityBase, exclude: list = [],
                           include_defaults: bool = True) -> dict:
    """Automatically match `obj` object data to function arguments.

    Handles default parameters. Extracts value from Enums. Returns matched
    arguments as dict.

    Args:
        function: function for which args will be matched
        obj: object to use for matching the function args
        exclude: set `exclude` parameter to exclude specific param-value pairs
        include_defaults: if `False` then values which have the same value as
            default will not be included in the result
    Raises:
        KeyError: could not match all required arguments
    """
    # convert names starting with '_'
    obj_dict = {key[1:] if key.startswith("_") else key: val for key, val in obj.__dict__.items()}
    kwargs = helper.auto_match_args(func, obj_dict, exclude, include_defaults)

    if "object_type" in kwargs:
        kwargs.update({"object_type": obj._OBJECT_TYPE.value})

    return kwargs



from enum import Enum
from typing import TYPE_CHECKING

from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.connection import Connection


class ShowExpressionAs(Enum):
    TEXT = None,
    TREE = "tree",
    TOKENS = "tokens"


class UpdateOperator(Enum):
    APPLY = "addElements"
    REVOKE = "removeElements"


@ErrorHandler(err_msg='Error getting members of security filter with ID {id}')
def get_security_filter_members(connection: "Connection", id: str, project_id: str = None,
                                error_msg: str = None):
    """Get the users and user groups that the specified security filter is
    applied to."""
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id
    return connection.session.get(
        url=f"{connection.base_url}/api/securityFilters/{id}/members",
        headers={'X-MSTR-ProjectID': project_id}
    )


@ErrorHandler(err_msg='Error updating members of security filter with ID {id}')
def update_security_filter_members(connection: "Connection", id: str, body: dict,
                                   project_id: str = None, error_msg: str = None,
                                   throw_error: bool = True):
    """Update members information for a specific security filter."""
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id
    return connection.session.patch(
        url=f"{connection.base_url}/api/securityFilters/{id}/members",
        headers={'X-MSTR-ProjectID': project_id},
        json=body
    )


@ErrorHandler(err_msg='Error creating new security filter.')
def create_security_filter(connection: "Connection", changeset_id: str, body: dict,
                           show_filter_tokens: bool = False, show_expression_as: str = None,
                           error_msg: str = None, throw_error: bool = True, **kwargs):
    """Create a security filter."""
    return connection.session.post(
        url=f"{connection.base_url}/api/model/securityFilters",
        headers={'X-MSTR-MS-Changeset': changeset_id},
        json=body,
        params={
            'showFilterTokens': str(show_filter_tokens).lower(),
            'showExpressionAs': show_expression_as
        }
    )


@ErrorHandler(err_msg='Error getting security filter {id} definition.')
def read_security_filter(connection: "Connection", id: str, project_id: str = None,
                         changeset_id: str = None, show_expression_as: str = None,
                         show_fields: str = None, show_filter_tokens: bool = False,
                         error_msg: str = None):
    """Read security filter."""
    return connection.session.get(
        url=f"{connection.base_url}/api/model/securityFilters/{id}",
        headers={
            'X-MSTR-ProjectID': project_id,
            'X-MSTR-MS-Changeset': changeset_id
        },
        params={
            'showExpressionAs': show_expression_as,
            'showFields': show_fields,
            'showFilterTokens': str(show_filter_tokens).lower()
        }
    )


@ErrorHandler(err_msg='Error updating security filter with ID {id}')
def update_security_filter(connection: "Connection", id: str, changeset_id: str, body: dict,
                           show_expression_as: str = None, show_fields: str = None,
                           show_filter_tokens: bool = False, error_msg: str = None,
                           throw_error: bool = True):
    """Update security filter."""
    return connection.session.put(
        url=f"{connection.base_url}/api/model/securityFilters/{id}",
        headers={'X-MSTR-MS-Changeset': changeset_id},
        params={
            'showExpressionAs': show_expression_as,
            'showFields': show_fields,
            'showFilterTokens': str(show_filter_tokens).lower()
        },
        json=body
    )



from enum import IntEnum

from mstrio.object_management import full_search as list_objects  # noqa: F401
from mstrio.utils.helper import deprecation_warning  # noqa: F401

deprecation_warning("mstrio.browsing.browsing", "mstrio.object_management.search_operations",
                    "11.3.4.101")  # NOSONAR


class SearchType(IntEnum):
    """Enumeration constants used to specify searchType used to control BI
    Search. More details can be found in EnumDSSXMLSearchTypes in a browser."""
    BEGIN_WITH = 1,
    BEGIN_WITH_PHRASE = 3,
    CONTAINS = 4,
    CONTAINS_ANY_WORD = 0,
    END_WITH = 5,
    EXACTLY = 2,


class SearchDomain(IntEnum):
    """Enumeration constants used to specify the search domains. More details
     can be found in EnumDSSXMLSearchDomain in a browser."""
    CONFIGURATION_AND_ALL_PROJECTS = 5,
    DOMAIN_CONFIGURATION = 4,
    DOMAIN_LOCAL = 1,
    DOMAIN_PROJECT = 2,
    DOMAIN_REPOSITORY = 3,



import sys
from enum import Enum
from typing import Union, List, TYPE_CHECKING

from mstrio.access_and_security.security_role import SecurityRole  # noqa: F401
from mstrio.access_and_security.security_filter import SecurityFilter  # noqa: F401
from mstrio.project_objects import Document, Report  # noqa: F401
from mstrio.project_objects.datasets import OlapCube, SuperCube  # noqa: F401
from mstrio.users_and_groups import User, UserGroup  # noqa: F401
from mstrio.server import Project  # noqa: F401
from mstrio.distribution_services.schedule import Schedule  # noqa: F401
from mstrio.datasources import DatasourceInstance, DatasourceLogin,\
    DatasourceConnection  # noqa: F401
from mstrio.object_management.folder import Folder  # noqa: F401
from mstrio.object_management.object import Object
from mstrio.object_management.search_operations import SearchObject  # noqa: F401

from mstrio.types import ObjectTypes, ObjectSubTypes

if TYPE_CHECKING:
    from mstrio.connection import Connection


class TypeObjectMapping(Enum):
    Report = ObjectTypes.REPORT_DEFINITION  # noqa: F811
    Folder = ObjectTypes.FOLDER  # noqa: F811
    DatasourceInstance = ObjectTypes.DBROLE  # noqa: F811
    DatasourceLogin = ObjectTypes.DBLOGIN  # noqa: F811
    DatasourceConnection = ObjectTypes.DBCONNECTION  # noqa: F811
    Project = ObjectTypes.PROJECT  # noqa: F811
    SecurityRole = ObjectTypes.SECURITY_ROLE  # noqa: F811
    Schedule = ObjectTypes.SCHEDULE_TRIGGER  # noqa: F811
    Document = ObjectTypes.DOCUMENT_DEFINITION  # noqa: F811
    SecurityFilter = ObjectTypes.SECURITY_FILTER  # noqa: F811
    SearchObject = ObjectTypes.SEARCH  # noqa: F811


class SubTypeObjectMapping(Enum):
    OlapCube = ObjectSubTypes.OLAP_CUBE  # noqa: F811
    SuperCube = ObjectSubTypes.SUPER_CUBE  # noqa: F811
    User = ObjectSubTypes.USER  # noqa: F811
    UserGroup = ObjectSubTypes.USER_GROUP  # noqa: F811


def __str_to_class(classname):
    return getattr(sys.modules[__name__], classname)


def map_to_object(object_type: Union[int, ObjectTypes], subtype: Union[int,
                                                                       ObjectSubTypes] = None):
    if not isinstance(object_type, ObjectTypes):
        try:
            object_type = ObjectTypes(object_type)
        except ValueError:
            object_type = None
    if not isinstance(subtype, ObjectSubTypes):
        try:
            subtype = ObjectSubTypes(subtype)
        except ValueError:
            subtype = None

    if ((object_type == ObjectTypes.REPORT_DEFINITION
         and subtype in (ObjectSubTypes.OLAP_CUBE, ObjectSubTypes.SUPER_CUBE))
            or (object_type == ObjectTypes.USER
                and subtype in (ObjectSubTypes.USER, ObjectSubTypes.USER_GROUP))):
        return __str_to_class(SubTypeObjectMapping(subtype).name)
    elif object_type in [v.value for v in TypeObjectMapping]:
        return __str_to_class(TypeObjectMapping(object_type).name)
    else:
        return Object


def map_objects_list(connection: "Connection", objects_list: List):
    return [
        map_to_object(obj.get('type'), obj.get('subtype')).from_dict(source=obj,
                                                                     connection=connection)
        for obj in objects_list
    ]



import os

from setuptools import dist, find_packages, setup
from setuptools.extern.packaging import version as packaging_version

from mstrio import (__author__, __author_email__, __description__, __license__, __title__,
                    __version__)

MANIFEST_FILE = 'MANIFEST.in'

if 'version.txt' in os.listdir():
    with open('version.txt') as f:
        dist_version = f.read().strip()
else:
    dist_version = __version__  # define the default version

# Patch Version class to preserve original version string
dist.pkg_resources.safe_version = lambda v: v


class NoNormalizeVersion(packaging_version.Version):

    def __init__(self, version):
        self._orig_version = version
        super().__init__(version)

    def __str__(self):
        return self._orig_version


def find_in_file(text, file):
    if file in os.listdir():
        with open(file, 'r') as f:
            manifest = f.read()
        return text in manifest
    else:
        raise ValueError(f'Could not find file {file}')


packaging_version.Version = NoNormalizeVersion
with open('README.md') as f:
    long_description = f.read()

requirements = [
    'requests>=2.25, <2.26',
    'urllib3>=1.26.0',
    'requests_futures>=1.0.0, <1.1',
    'pandas>=1.1.5, <1.3',
    'numpy>=1.18.1, <1.21',
    'tqdm>=4.41, <4.70',
    'packaging>=20.9, <21',
    'dictdiffer>=0.8.1, <0.9',
    'stringcase>=1.2, <1.3',
    'Jinja2>=2.11, <3.0',
]

# Add dependencies for connector-jupyter if connector-jupyter folder is added
# Else add workstation_demo_scripts folder
if find_in_file('graft connector-jupyter', MANIFEST_FILE):
    requirements.extend(['jupyter_contrib_nbextensions>=0.5.1, <0.6', 'ipywidgets>=7.5.1, <8'])
elif find_in_file('prune connector-jupyter', MANIFEST_FILE):
    with open(MANIFEST_FILE, "a+") as f:
        f.write('graft workstation_demo_scripts')

setup(
    name=__title__,
    python_requires='>=3.6',
    version=dist_version,
    license=__license__,
    description=__description__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/MicroStrategy/mstrio-py',
    author=__author__,
    author_email=__author_email__,
    project_urls={
        'Bug Tracker': 'https://github.com/MicroStrategy/mstrio-py/issues',
        'Documentation': 'http://www2.microstrategy.com/producthelp/Current/mstrio-py/',
        'Source Code': 'https://github.com/MicroStrategy/mstrio-py',
        'Quick Manual': 'https://www2.microstrategy.com/producthelp/current/MSTR-for-Jupyter/Content/mstr_for_jupyter.htm',  # noqa
    },
    install_requires=requirements,
    extras_require={
        'dev': ['flake8', 'mypy', 'yapf', 'nose', 'coverage', 'pytest', 'pytest-cov',
                'isort'],  # noqa
    },
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: JavaScript',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Libraries',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
    ],
)



from enum import Enum
from typing import Dict, List, Optional, TYPE_CHECKING, Union

from mstrio.access_and_security.security_filter import (AttributeElement, AttributeFormNormalRef,
                                                        AttributeFormSystemRef, AttributeRef,
                                                        ConstantArrayParameter, ConstantParameter,
                                                        DynamicDateTimeParameter, ElementPromptRef,
                                                        ExpressionParameter, FilterRef,
                                                        ObjectReferenceParameter, ParameterType,
                                                        PromptParameter)
from mstrio.utils.helper import Any, camel_to_snake, delete_none_values, Dictable, snake_to_camel

if TYPE_CHECKING:
    from mstrio.access_and_security.security_filter import ParameterBase
    from mstrio.connection import Connection


class PredicateBase(Dictable):

    TYPE = None

    def _create_predicate_tree(self, **kwargs):
        return None

    def _predicate_tree_to_dict(self, camel_case=True):
        return {}

    def to_dict(self, camel_case=True):
        result = delete_none_values({
            "type": self.type,
            "predicate_id": self.predicate_id,
            "predicate_tree": self._predicate_tree_to_dict(camel_case)
        })
        return snake_to_camel(result) if camel_case else result

    def __init__(self, predicate_id: Optional[str] = None, **kwargs):
        self.type = self.TYPE
        self.predicate_id = predicate_id
        self.predicate_tree = self._create_predicate_tree(**kwargs)


class PredicateFormFunction(Enum):
    EQUALS = "equals"
    NOT_EQUAL = "not_equals"
    GREATER = "greater"
    GREATER_EQUALS = "greater_equals"
    LESS = "less"
    LESS_EQUAL = "less_equal"
    IS_NULL = "is_null"
    IS_NOT_NULL = "is_not_null"
    BEGINS_WITH = "begins_with"
    NOT_BEGINS_WITH = "not_begins_with"
    CONTAINS = "contains"
    NOT_CONTAINS = "not_contains"
    LIKE = "like"
    NOT_LIKE = "not_like"
    IN = "in"
    NOT_IN = "not_in"
    BETWEEN = "between"
    NOT_BETWEEN = "not_between"


class PredicateForm(PredicateBase):
    # Attribute Form Predicate
    TYPE = 'predicate_form_qualification'

    def _create_predicate_tree(self, **kwargs):
        self.function = kwargs.get("function")
        self.parameters = kwargs.get("parameters")
        self.attribute = kwargs.get("attribute")
        self.form = kwargs.get("form")
        self.data_locale = kwargs.get("data_locale")

    def _predicate_tree_to_dict(self, camel_case=True):
        result = {
            "function": self.function.value,
            "parameters": [p.to_dict() for p in self.parameters],
            "attribute": self.attribute.to_dict(),
            "form": self.form.to_dict(),
            "dataLocale": self.data_locale
        }
        return snake_to_camel(result) if camel_case else result

    @classmethod
    def from_dict(cls, source: Dict[str, Any], connection: Optional["Connection"] = None):
        source = camel_to_snake(source)
        predicate_id = source.get("predicate_id")

        source = source.get('predicate_tree')
        function = PredicateFormFunction(source.get("function"))

        parameters = []
        for parameter in source.get("parameters", []):
            parameter_type = parameter.get("parameterType")
            if parameter_type == ParameterType.CONSTANT.value:
                parameters.append(ConstantParameter.from_dict(parameter, connection))
            elif parameter_type == ParameterType.OBJECT_REFERENCE.value:
                parameters.append(
                    ObjectReferenceParameter.from_dict(parameter, connection))
            elif parameter_type == ParameterType.EXPRESSION.value:
                parameters.append(
                    ExpressionParameter.from_dict(parameter, connection))
            elif parameter_type == ParameterType.PROMPT.value:
                parameters.append(PromptParameter.from_dict(parameter, connection))
            elif parameter_type == ParameterType.DYNAMIC_DATE_TIME.value:
                parameters.append(
                    DynamicDateTimeParameter.from_dict(parameter, connection))
            elif parameter_type == ParameterType.ARRAY.value:
                parameters.append(
                    ConstantArrayParameter.from_dict(parameter, connection))

        attribute = AttributeRef.from_dict(source.get("attribute"), connection)
        form = source.get("form", {})
        form_sub_type = form.get("sub_type")
        if form_sub_type == "attribute_form_system":
            form = AttributeFormSystemRef.from_dict(source.get("form"), connection)
        elif form_sub_type == "attribute_form_normal":
            form = AttributeFormNormalRef.from_dict(source.get("form"), connection)
        else:
            form = None
        data_locale = source.get("data_locale")

        return PredicateForm(function, parameters, attribute, form, data_locale, predicate_id)

    def __init__(self, function: "PredicateFormFunction", parameters: List["ParameterBase"],
                 attribute: "AttributeRef", form: Union["AttributeFormSystemRef",
                                                        "AttributeFormNormalRef"],
                 data_locale: Optional[str] = None, predicate_id: Optional[str] = None):
        """Specialized expression node that contains an attribute form
        qualification predicate.

        This qualification selects elements of a specific attribute by
        comparing the value of a specified form of the element with an
        expression. The comparison operator used is also specified. However
        there are three options for specifying the expression with which it is
        compared:
            - the client may specify a literal value
            - the client may ask to use a prompt to determine the value
            - the client may supply an arbitrary metric style expression to
              evaluate to obtain the value

        Args:
            function (enum): the comparison function that should be used to
                compare the value computed by the filter with an expression
                value. Available values are defined in enum
                `PredicateFormFunction`.
            parameters (list of objects): array of objects that correspond to
                the parameters for the comparison function. In nearly all cases
                the comparison function will take one extra parameter (in
                addition to the parameter referenced by the predicate's main
                object). This value is an array because there are a few
                functions which take a different number of extra parameters. For
                example `isNull` does not need any additional parameters,
                `notBetween` needs two additional parameters.
            attribute (object): object with basic information about attribute
                such as its `object_id` and `name`.
            form (object): object with basic information about attribute form
                such as its `object_id` and `name`.
            data_locale (str): optional data locale used to select value of the
                form.
            predicate_id (str): identifier of this predicate. When creating
                predicate it is not necessary to provide it.

        Example:
            >>> PredicateForm(
            >>>     function=PredicateFormFunction.GREATER,
            >>>     parameters=[
            >>>         ConstantParameter(ConstantType.DOUBLE, "2015.0")
            >>>     ],
            >>>     attribute=AttributeRef("8D679D5111D3E4981000E787EC6DE8A4",
            >>>                            "Year"),
            >>>     form=AttributeFormSystemRef(
            >>>         "45C11FA478E745FEA08D781CEA190FE5", "ID"
            >>>     )
            >>> )
        """
        super().__init__(predicate_id, function=function, parameters=parameters,
                         attribute=attribute, form=form, data_locale=data_locale)


class PredicateElementListFunction(Enum):
    NOT_IN = "notIn"
    IN = "in"


class PredicateElementList(PredicateBase):
    # Element List Predicate
    TYPE = 'predicate_element_list'

    def _create_predicate_tree(self, **kwargs):
        self.function = kwargs.get("function")
        self.attribute = kwargs.get("attribute")
        self.elements = kwargs.get("elements")
        self.elements_prompt = kwargs.get("elements_prompt")

    def _predicate_tree_to_dict(self, camel_case=True):
        elements_prompt = None if self.elements_prompt is None else [
            elem.to_dict(camel_case) for elem in self.elements_prompt
        ]
        result = {
            "function": self.function.value,
            "attribute": self.attribute.to_dict(camel_case),
            "elements": [elem.to_dict(camel_case) for elem in self.elements],
            "elements_prompt": elements_prompt
        }
        return snake_to_camel(result) if camel_case else result

    @classmethod
    def from_dict(cls, source: Dict[str, Any], connection: Optional["Connection"] = None):
        source = camel_to_snake(source)
        predicate_id = source.get("predicate_id")

        source = source.get('predicate_tree')
        function = PredicateElementListFunction(source.get("function"))
        attribute = AttributeRef.from_dict(source.get("attribute"), connection)
        elements = [
            AttributeElement.from_dict(elem, connection)
            for elem in source.get("elements")
        ]
        elements_prompt = None if source.get("elements_prompt") is None else [
            ElementPromptRef.from_dict(elem, connection)
            for elem in source.get("elements_prompt")
        ]
        return PredicateElementList(function, attribute, elements, elements_prompt, predicate_id)

    def __init__(self, function: "PredicateElementListFunction", attribute: "AttributeRef",
                 elements: List["AttributeElement"],
                 elements_prompt: Optional[List["ElementPromptRef"]] = None,
                 predicate_id: Optional[str] = None):
        """Specialized expression node that contains an element list predicate.

        This qualification selects elements of a specified attribute by either
        listing the elements that should be selected, or alternatively by
        listing the elements that should be excluded from the selection.

        Args:
            function (enum): function that should be used in this predicate.
                Available values are defined in enum
                `PredicateElementsListFunction`.
            attribute (object): object with basic information about attribute
                such as its `object_id` and `name`.
            elements (list of objects): list of objects with basic information
                about element such its `element_id` and `display`.
            elements_prompt (list of objects): list of objects with basic
                information about a prompt element such its `object_id` and
                `name`
            predicate_id (str): identifier of this predicate. When creating
                predicate it is not necessary to provide it.

        Example:
            >>> PredicateElementList(
            >>>     function=PredicateElementListFunction.IN,
            >>>     attribute=AttributeRef("8D679D3711D3E4981000E787EC6DE8A4",
            >>>                            "Catetgory"),
            >>>     elements=[
            >>>         AttributeElement("h2", "Electronics"),
            >>>         AttributeElement("h1", "Books")
            >>>     ]
            >>> )
        """
        super().__init__(predicate_id, function=function, attribute=attribute, elements=elements,
                         elements_prompt=elements_prompt)


class PredicateFilter(PredicateBase):
    # Filter Qualification Predicate
    TYPE = 'predicate_filter_qualification'

    def _create_predicate_tree(self, **kwargs):
        self.filter = kwargs.get("filter")
        self.is_independent = kwargs.get("is_independent")

    def _predicate_tree_to_dict(self, camel_case=True):
        result = {
            "filter": self.filter.to_dict(camel_case),
            "is_independent": 1 if self.is_independent else 0,
        }
        return snake_to_camel(result) if camel_case else result

    @classmethod
    def from_dict(cls, source: Dict[str, Any], connection: Optional["Connection"] = None):
        source = camel_to_snake(source)
        predicate_id = source.get("predicate_id")

        source = source.get('predicate_tree')
        filter = FilterRef.from_dict(source.get("filter"), connection)
        is_independent = bool(source.get("is_independent"))

        return PredicateFilter(filter, is_independent, predicate_id)

    def __init__(self, filter: "FilterRef", is_independent: Optional[bool] = True,
                 predicate_id: Optional[str] = None):
        """Specialized expression node that contains a filter qualification
        predicate.

        This node is used within filter expressions to represent
        a predicate whose value is determined by using an external filter
        object. At execution time the engine will typically act as if the
        specified filter had been included directly. By using filter objects,
        the client can build complex filter expressions that utilise common
        subexpressions.

        Args:
            filter (object): object with basic information about child filter
                such as its `id` and `name`
            is_independent (bool): flag that indicates whethere this child will
                be considered independently of other parts of the larger filter.
                Default value is `True` which means that this filter will be
                evaluated by itself.
            predicate_id (str): identifier of this predicate. When creating
                predicate it is not necessary to provide it.

        Example:
            >>> PredicateFilter (
            >>>     filter=FilterRef("320081BF47ECD3DEB07529B1BEF4271B",
            >>>                      "Filter Object"),
            >>>     is_independent=True
            >>> )
        """
        super().__init__(predicate_id, filter=filter, is_independent=is_independent)


class PredicateJointElementList(PredicateBase):
    # Joint Element List Predicate
    TYPE = 'predicate_joint_element_list'

    def _create_predicate_tree(self, **kwargs):
        self.level = kwargs.get("level")
        self.tuples = kwargs.get("tuples")

    def _predicate_tree_to_dict(self, camel_case=True):
        result = {
            "level": [obj.to_dict(camel_case) for obj in self.level],
            "tuples": [[elem.to_dict(camel_case) for elem in obj] for obj in self.tuples],
            "predicate_id": self.predicate_id
        }
        return snake_to_camel(result) if camel_case else result

    @classmethod
    def from_dict(cls, source: Dict[str, Any], connection: Optional["Connection"] = None):
        source = camel_to_snake(source)
        predicate_id = source.get("predicate_id")

        source = source.get('predicate_tree')
        level = [
            AttributeRef.from_dict(obj, connection) for obj in source.get("level")
        ]
        tuples = [[AttributeElement.from_dict(elem, connection)
                   for elem in obj]
                  for obj in source.get("tuples", [])]

        return PredicateJointElementList(predicate_id, level, tuples)

    def __init__(self, level: List["AttributeRef"], tuples: List[List["AttributeElement"]],
                 predicate_id: Optional[str] = None):
        """Specialized expression node that contains a joint element list
        predicate.

        This qualification represents a filter at the level of two or more
        attributes. The filter is defined by simply listing the tuples of
        elements that satisfy the filter.

        Args:
            level (list of objects): the level of this filter is represented as
                an array of attributes. The attributes must be different from
                each other, and their order is significant.
            tuples (list of lists of objects): an array of tuples of elements.
                Each element of this array is a joint element. That is the
                members of this array are themselves arrays of elements. The
                elements in the tuples must be listed in the same order as the
                attributtes were listed in the `level` array. There is no need
                to list the attribute on each element.
            predicate_id (str): identifier of this predicate. When creating
                predicate it is not necessary to provide it.

        Example:
            >>> PredicateJointElementList(
            >>>     level=[
            >>>         AttributeRef("8D679D3711D3E4981000E787EC6DE8A4",
            >>>                      "Category"),
            >>>         AttributeRef("8D679D4B11D3E4981000E787EC6DE8A4",
            >>>                      "Region")
            >>>     ],
            >>>     tuples=[
            >>>         [AttributeElement("h1", "Books"),
            >>>          AttributeElement("h1", "Northeast")],
            >>>         [AttributeElement("h2", "Electronics"),
            >>>          AttributeElement("h7", "Southwest")]
            >>>     ])
        """
        super().__init__(predicate_id, level=level, tuples=tuples)


class LogicFunction(Enum):
    OR = "or"
    AND = "and"


class LogicOperator(Dictable):

    TYPE = "operator"

    def __init__(self, function: "LogicFunction", children: Union["LogicOperator",
                                                                  "PredicateBase"]):
        self.type = "operator"
        self.function = function
        self.children = children

    def to_dict(self, camel_case=True):
        result = delete_none_values({
            "type": self.type,
            "function": self.function.value,
            "children": [child.to_dict(camel_case) for child in self.children]
        })
        return snake_to_camel(result) if camel_case else result

    @classmethod
    def from_dict(cls, source: Dict[str, Any], connection: Optional["Connection"] = None):
        source = camel_to_snake(source)
        function = LogicFunction(source.get("function"))
        children = []
        for child in source.get("children"):
            child_type = child.get("type")
            if child_type == "predicate_form_qualification":
                children.append(PredicateForm.from_dict(child, connection))
            elif child_type == "predicate_joint_element_list":
                children.append(
                    PredicateJointElementList.from_dict(child, connection))
            elif child_type == "predicate_filter_qualification":
                children.append(PredicateFilter.from_dict(child, connection))
            elif child_type == "predicate_element_list":
                children.append(PredicateElementList.from_dict(child, connection))
            elif child_type == "operator":
                children.append(LogicOperator.from_dict(child, connection))

        return LogicOperator(function, children)



from typing import TYPE_CHECKING

from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from requests_futures.sessions import FuturesSession


@ErrorHandler(err_msg='Error getting available dossiers.')
def get_dossiers(connection, search_term, certified_status, search_pattern='EXACTLY', offset=0,
                 limit=-1, fields=None, error_msg=None):
    """Get the list of available dossiers.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f'{connection.base_url}/api/dossiers'
    return connection.session.get(
        url=url,
        params={
            'searchTerm': search_term,
            'searchPattern': search_pattern,
            'offset': offset,
            'limit': limit,
            'fields': fields
        },
    )


def get_dossiers_async(future_session: "FuturesSession", connection, search_term, certified_status,
                       search_pattern='EXACTLY', offset=0, limit=-1, fields=None, error_msg=None):
    """Get the list of available dossiers asynchronously.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    connection._validate_project_selected()
    url = f'{connection.base_url}/api/dossiers'
    params = {
        'searchTerm': search_term,
        'searchPattern': search_pattern,
        'offset': offset,
        'limit': limit,
        'fields': fields
    }
    return future_session.get(url=url, params=params)


@ErrorHandler(err_msg='Error getting available documents.')
def get_documents(connection, search_term, certified_status, search_pattern='EXACTLY', offset=0,
                  limit=-1, fields=None, error_msg=None):
    """Get the list of available documents.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f'{connection.base_url}/api/documents/'
    return connection.session.get(
        url=url,
        params={
            'searchTerm': search_term,
            'searchPattern': search_pattern,
            'offset': offset,
            'limit': limit,
            'fields': fields
        },
    )


def get_documents_async(future_session: "FuturesSession", connection, search_term,
                        certified_status, search_pattern='EXACTLY', offset=0, limit=-1,
                        fields=None, error_msg=None):
    """Get the list of available documents asynchronously.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f'{connection.base_url}/api/documents/'
    params = {
        'searchTerm': search_term,
        'searchPattern': search_pattern,
        'offset': offset,
        'limit': limit,
        'fields': fields
    }
    future = future_session.get(url=url, params=params)
    return future


@ErrorHandler(err_msg='Error getting document {document_id}')
def get_document_status(connection, document_id, instance_id, error_msg=None):
    """Get the status of a document or dossier instance.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/documents​/{document_id}​/instances​/{instance_id}​/status"
    return connection.session.get(url=url, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg='Error getting prompts for document {document_id}')
def get_prompts_for_instance(connection, document_id, instance_id, error_msg=None):
    """Get the collection of prompts and their respective definitions from a
    document/dossier instance.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/documents​/{document_id}​/instances​/{instance_id}​/prompts"  # noqa
    return connection.session.get(url=url, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg='Error getting attribute element for prompt {prompt_identifier}')
def get_attribute_element_for_prompt(connection, document_id, instance_id, prompt_identifier,
                                     error_msg=None):
    """Get available attribute element for document/dossier's attribute element
    prompt.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        prompt_identifier (string): Prompt key or ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/documents​/{document_id}​/instances​/{instance_id}​/prompts​/{prompt_identifier}​/elements"  # noqa
    return connection.session.get(url=url, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg='Error getting available object for prompt {prompt_identifier}')
def get_available_object(connection, document_id, instance_id, prompt_identifier, error_msg=None):
    """Get available object for answering all kinds of prompts.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        prompt_identifier (string): Prompt key or ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}/api​/documents​/{document_id}​/instances​/{instance_id}​/prompts​/{prompt_identifier}​/objects"  # noqa
    return connection.session.get(url=url, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg='Error exporting visualization for document {document_id} to PDF file.')
def export_visualization_to_pdf(connection, document_id, instance_id, node_key, body,
                                error_msg=None):
    """Export a single visualization from a specific document instance to a PDF
    file.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/documents​/{document_id}​/instances​/{instance_id}​/visualizations​/{node_key}​/pdf"  # noqa
    return connection.session.post(url=url, headers={'X-MSTR-ProjectID': None}, json=body)


@ErrorHandler(err_msg='Error exporting visualization for document {document_id} to CSV file.')
def export_visualization_to_csv(connection, document_id, instance_id, node_key, body,
                                error_msg=None):
    """Export a single visualization from a specific document instance to a CSV
    file.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        node_key (string): Visualization node key
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/documents​/{document_id}​/instances​/{instance_id}​/visualizations​/{node_key}​/csv"  # noqa
    return connection.session.post(url=url, headers={'X-MSTR-ProjectID': None}, json=body)


@ErrorHandler(err_msg='Error exporting document {document_id} to PDF file.')
def export_document_to_pdf(connection, document_id, instance_id, body, error_msg=None):
    """Export a specific document instance to a PDF file.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/documents​/{document_id}​/instances​/{instance_id}​/pdf"
    return connection.session.post(url=url, headers={'X-MSTR-ProjectID': None}, json=body)


@ErrorHandler(err_msg='Error exporting document {document_id} to .mstr file')
def export_document_to_mstr(connection, document_id, instance_id, body, error_msg=None):
    """Export a specific document in a specific project to an .mstr file.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/documents​/{document_id}​/instances​/{instance_id}​/mstr"
    return connection.session.post(url=url, headers={'X-MSTR-ProjectID': None}, json=body)


@ErrorHandler(err_msg='Error exporting document {document_id} to Excel file.')
def export_document_to_excel(connection, document_id, instance_id, body, error_msg=None):
    """Export a document from a specific document instance to an Excel file.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/documents​/{document_id}​/instances​/{instance_id}​/excel"
    return connection.session.post(url=url, headers={'X-MSTR-ProjectID': None}, json=body)


@ErrorHandler(err_msg='Error setting document {document_id} to prompt status.')
def set_document_to_prompt_status(connection, document_id, instance_id, error_msg=None):
    """Export a document from a specific document instance to an Excel file.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/documents​/{document_id}​/instances​/{instance_id}​/rePrompt"  # noqa
    return connection.session.post(url=url, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg='Error getting cubes used by document {document_id}')
def get_cubes_used_by_document(connection, document_id, error_msg=None):
    """Get the cubes used by a document in a specific project, either directly
    or indirectly.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}/api/documents/{document_id}/cubes"
    return connection.session.get(url=url)


@ErrorHandler(err_msg='Error overwriting document {document_id}')
def overwrite_document(connection, document_id, instance_id, error_msg=None):
    """Save a document instance by overwriting an existing document.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/documents​/{document_id}​/instances​/{instance_id}​/save"
    return connection.session.post(url=url, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg='Error saving document {document_id}')
def save_document_as(connection, document_id, instance_id, error_msg=None):
    """Save a document instance by creating a new document.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}/api​/documents​/{document_id}​/instances​/{instance_id}​/saveAs"  # noqa
    return connection.session.post(url=url, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg='Error creating instance for document {document_id}')
def create_new_document_instance(connection, document_id, body, error_msg=None):
    """Execute a specific document in a specific project and create an instance
    of the document.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}/api/documents/{document_id}/instances"
    return connection.session.post(url=url, json=body)


@ErrorHandler(err_msg='Error deleting document instance {instance_id}')
def delete_document_instance(connection, document_id, instance_id, error_msg=None):
    """Save a document instance by overwriting an existing document.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/documents​/{document_id}​/instances​/{instance_id}​"
    return connection.session.delete(url=url, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg='Error refreshing document instance {instance_id}')
def refresh_document_instance(connection, document_id, instance_id, error_msg=None):
    """Execute a specific document in a specific project and create an instance
    of the document.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/documents​/{document_id}​/instances/{instance_id}/refresh​"
    return connection.session.put(url=url, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg='Error getting collection of prompts for document {document_id}')
def get_prompts(connection, document_id, error_msg=None):
    """Get the collection of prompts and their respective definitions from a
    document/dossier definition.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/documents​/{document_id}​/prompts"
    return connection.session.get(url=url, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg='Error answering prompt for instance {instance_id}')
def answer_prompts(connection, document_id, instance_id, body, error_msg=None):
    """Answer specified prompts on the document/dossier instance, prompts can
    either be answered with default answers(if available), the appropriate
    answers, or if the prompt is not required the prompt can simply be closed.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}/api​/documents​/{document_id}​/instances/{instance_id}/prompts/answers"  # noqa
    return connection.session.put(url=url, headers={'X-MSTR-ProjectID': None}, json=body)


@ErrorHandler(err_msg='Error retrieving document shortcut for document {document_id}')
def get_document_shortcut(connection, document_id, instance_id, error_msg=None):
    """Retrieve a published shortcut from a specific document instance.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}/api​/documents​/{document_id}​/instances/{instance_id}/shortcut"
    return connection.session.get(url=url)


@ErrorHandler(err_msg='Error creating instance for dossier {dossier_id}')
def create_dossier_instance(connection, dossier_id, body, error_msg=None):
    """Execute a specific dossier and create an instance of the dossier.

    Args:
        connection: MicroStrategy REST API connection object
        dossier_id (string): Dossier ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}​/api​/dossiers/{dossier_id}​/instances​"
    return connection.session.post(url=url, headers={'X-MSTR-ProjectID': None}, json=body)


@ErrorHandler(err_msg='Error getting hierarchy for dossier {id}')
def get_dossier_hierarchy(connection, id, error_msg=None):
    """Get the hierarchy of a specific dossier in a specific project.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Dossier ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}/api/v2/dossiers/{id}/definition/"
    return connection.session.get(url=url, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg='Error getting definition of document {id}')
def get_document_definition(connection, id, error_msg=None):
    """Get the hierarchy of a specific dossier in a specific project.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Dossier ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}/api/v2/documents/{id}"
    return connection.session.get(url=url, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg='Error getting dossier hierarchy from instance {instance_id}')
def get_dossier_hierarchy_from_instance(connection, dossier_id, instance_id, error_msg=None):
    """Get the hierarchy of a specific dossier in a specific project from
    instance.

    Args:
        connection: MicroStrategy REST API connection object
        dossier_id (string): Dossier ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}/api​/v2/dossiers/{dossier_id}/​instances/{instance_id}/definition"  # noqa
    return connection.session.get(url=url, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg='Error getting definition and results for dossier {dossier_id}')
def get_definition_and_results_of_visualization(connection, dossier_id, instance_id, chapter_key,
                                                visualization_key, error_msg=None):
    """Get the definition and data result of a grid/graph visualization in a
    specific dossier in a specific project.

    Args:
        connection: MicroStrategy REST API connection object
        dossier_id (string): Dossier ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}/api/​v2/dossiers​/{dossier_id}​/instances​/{instance_id}​/chapters​/{chapter_key}​/visualizations​/{visualization_key}"  # noqa
    return connection.session.get(url=url, headers={'X-MSTR-ProjectID': None})



from typing import Optional
from mstrio.utils.helper import Dictable


class ObjectInformation(Dictable):
    """An object that contains all of the type-neutral information stored in the
    metadata about an object.

    Attributes:
        id (string): a globally unique identifier used to distinguish between
            metadata objects within the same project.
        subtype (string): string literal used to identify the type of a metadata
            object.
        name (string): the name of the metadata object.
        is_embedded (boolean): True indicates that the target object of this
            reference is embedded within this object. Alternatively if this
            object is itself embedded, then it means that the target object is
            embedded in the same container as this object.
        description (string): user supplied description of this object.
        date_created (string): the date/time at which this object was first
            saved into the metadata.
        date_modified (string): the date/time at which this object was last
            saved into the metadata.
        destination_folder_id (string): id of a folder in which the object is
            stored.
        version_id (string): the version number this object is currently
            carrying.
        path (string): the path of this object
        acl (list of objects): an array of access control entry objects. Taken
            together these objects specify which users are permitted to perform
            which actions on this object.
        primary (string): the primary locale of the object, in the IETF BCP 47
            language tag format, such as "en-US".
    """

    def __init__(self, object_id: Optional[str] = None, sub_type: Optional[str] = None,
                 name: Optional[str] = None, is_embedded: Optional[bool] = None,
                 description: Optional[str] = None, date_created: Optional[str] = None,
                 date_modified: Optional[str] = None, destination_folder_id: Optional[str] = None,
                 version_id: Optional[str] = None, path: Optional[str] = None,
                 acl: Optional[dict] = None, primary_locale: Optional[str] = None):
        self.object_id = object_id
        self.sub_type = sub_type
        self.name = name
        self.is_embedded = is_embedded
        self.description = description
        self.date_created = date_created
        self.date_modified = date_modified
        self.destination_folder_id = destination_folder_id
        self.version_id = version_id
        self.path = path
        self.acl = acl
        self.primary_locale = primary_locale


class ObjectReference(Dictable):
    """Information about an object referenced within the specification of
    another object. An object reference typically contains only enough fields
    to uniquely identify the referenced objects.

    object_id (string): a globally unique identifier used to distinguish between
        metadata objects within the same project.
    sub_type (string): string literal used to identify the type of a metadata
        object
    name (string): the name of the metadata object.
    is_embedded (boolean): True indicates that the target object of this
        reference is embedded within this object. Alternatively if this
        object is itself embedded, then it means that the target object is
        embedded in the same container as this object.
    """

    def __init__(self, object_id: str, sub_type: str, name: Optional[str] = None,
                 is_embedded: Optional[bool] = None):
        self.object_id = object_id
        self.sub_type = sub_type
        self.name = name
        self.is_embedded = is_embedded


class _ObjectReferenceWithType(ObjectReference):
    SUB_TYPE = ""

    def __init__(self, object_id: str, name: Optional[str] = None,
                 is_embedded: Optional[bool] = None):
        super().__init__(object_id, self.SUB_TYPE, name, is_embedded)


class AttributeRef(_ObjectReferenceWithType):

    SUB_TYPE = "attribute"


class AttributeFormSystemRef(_ObjectReferenceWithType):

    SUB_TYPE = "attribute_form_system"


class AttributeFormNormalRef(_ObjectReferenceWithType):

    SUB_TYPE = "attribute_form_normal"


class AttributeElement(Dictable):

    def __init__(self, element_id: str, display: str, attribute: Optional[AttributeRef] = None):
        self.element_id = element_id
        self.display = display
        self.attribute = attribute


class ElementPromptRef(_ObjectReferenceWithType):

    SUB_TYPE = "prompt"


class FilterRef(_ObjectReferenceWithType):

    SUB_TYPE = "filter"



from typing import List, Optional, Union

from packaging import version

from mstrio import config
import mstrio.api.subscriptions as subscriptions_
from mstrio.connection import Connection
from mstrio.utils import helper

from . import CacheUpdateSubscription, EmailSubscription, Subscription
from .content import Content
from .delivery import Delivery


def list_subscriptions(connection: Connection, project_id: Optional[str] = None,
                       project_name: Optional[str] = None, to_dictionary: bool = False,
                       limit: Optional[int] = None,
                       **filters) -> Union[List["Subscription"], List[dict]]:
    """Get all subscriptions per project as list of Subscription objects or
    dictionaries.

    Optionally filter the subscriptions by specifying filters.
    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is
    omitted.

    Args:
        connection(object): MicroStrategy connection object
        project_id: Project ID
        project_name: Project name
        to_dictionary: If True returns a list of subscription dicts,
            otherwise (default) returns a list of subscription objects
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'editable',
            'allowDeliveryChanges', 'allowPersonalizationChanges',
            'allowUnsubscribe', 'dateCreated', 'dateModified', 'owner',
            'schedules', 'contents', 'recipients', 'delivery']
    """
    project_id = Subscription._project_id_check(connection, project_id, project_name)
    chunk_size = 1000 if version.parse(
        connection.iserver_version) >= version.parse('11.3.0300') else 1000000
    msg = 'Error getting subscription list.'
    objects = helper.fetch_objects_async(
        connection=connection,
        api=subscriptions_.list_subscriptions,
        async_api=subscriptions_.list_subscriptions_async,
        limit=limit,
        chunk_size=chunk_size,
        filters=filters,
        error_msg=msg,
        dict_unpack_value="subscriptions",
        project_id=project_id,
    )

    if to_dictionary:
        return objects
    else:
        return [
            dispatch_from_dict(
                source=obj,
                connection=connection,
                project_id=project_id,
            ) for obj in objects
        ]


DeliveryMode = Delivery.DeliveryMode
subscription_type_from_delivery_mode_dict = {
    DeliveryMode.CACHE: CacheUpdateSubscription,
    DeliveryMode.EMAIL: EmailSubscription,
}


def get_subscription_type_from_delivery_mode(mode: DeliveryMode):
    return subscription_type_from_delivery_mode_dict.get(mode, Subscription)


def dispatch_from_dict(source: dict, connection: Connection, project_id: str):
    delivery_mode = DeliveryMode(source["delivery"]["mode"])
    sub_type = get_subscription_type_from_delivery_mode(delivery_mode)
    return sub_type.from_dict(source, connection, project_id)


class SubscriptionManager:
    """Manage subscriptions."""

    def __init__(self, connection: Connection, project_id: Optional[str] = None,
                 project_name: Optional[str] = None, application_id: Optional[str] = None,
                 application_name: Optional[str] = None):
        """Initialize the SubscriptionManager object.
        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is
        omitted.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`
            project_id: Project ID
            project_name: Project name
            application_id: deprecated. Use project_id instead.
            application_name: deprecated. Use project_name instead.
        """
        if application_id or application_name:
            helper.deprecation_warning(
                '`application_name` and `application_id`',
                '`project_name` and `project_id`',
                '11.3.4.101',  # NOSONAR
                False)
            project_id = project_id or application_id
            project_name = project_name or application_name
        self.connection = connection
        self.project_id = Subscription._project_id_check(connection, project_id, project_name)

    def list_subscriptions(self, to_dictionary: bool = False, limit: Optional[int] = None,
                           **filters):
        """Get all subscriptions as list of Subscription objects or
        dictionaries.

        Optionally filter the subscriptions by specifying filters.

        Args:
            to_dictionary: If True returns a list of subscription dicts,
                otherwise returns a list of subscription objects
            limit: limit the number of elements returned. If `None` (default),
                all objects are returned.
            **filters: Available filter parameters: ['id', 'name', 'editable',
                'allowDeliveryChanges', 'allowPersonalizationChanges',
                'allowUnsubscribe', 'dateCreated', 'dateModified', 'owner',
                'schedules', 'contents', 'recipients', 'delivery']
        """
        return list_subscriptions(connection=self.connection, project_id=self.project_id,
                                  to_dictionary=to_dictionary, **filters)

    def delete(self, subscriptions: Union[List[Subscription], List[str]], force=False) -> bool:
        """Deletes all passed subscriptions. Returns True if successfully
        removed all subscriptions.

        Args:
            subscriptions: list of subscriptions to be deleted
        """
        subscriptions = subscriptions if isinstance(subscriptions, list) else [subscriptions]
        if not subscriptions and config.verbose:
            print("No subscriptions passed.")
        else:
            temp_subs = []
            for subscription in subscriptions:
                subscription = subscription if isinstance(
                    subscription, Subscription) else Subscription(self.connection, subscription,
                                                                  self.project_id)
                temp_subs.append(subscription)
            subscriptions = temp_subs
            succeeded = 0
            user_input = 'N'
            if not force:
                to_be_deleted = [
                    "Subscription '{}' with ID: '{}'".format(sub.name, sub.id)
                    for sub in subscriptions
                ]
                print("Found subscriptions:")
                for sub in to_be_deleted:
                    print(sub)
                user_input = input("Are you sure you want to delete all of them? [Y/N]: ")
            if force or user_input == 'Y':
                succeeded = 0
                for subscription in subscriptions:
                    response = subscriptions_.remove_subscription(
                        self.connection,
                        subscription.id,
                        self.project_id,
                        error_msg="Subscription '{}' with id '{}'' could not be deleted.".format(
                            subscription.name, subscription.id),
                        exception_type=UserWarning,
                    )
                    if response.ok:
                        succeeded += 1
                        if config.verbose:
                            print("Deleted subscription '{}' with ID '{}''.".format(
                                subscription.name, subscription.id))

                return succeeded == len(subscriptions)

    def execute(self, subscriptions: Union[List[Subscription], List[str]]):
        """Executes all passed subscriptions.

        Args:
            subscriptions: list of subscriptions to be executed
        """
        if not subscriptions and config.verbose:
            print("No subscriptions passed.")
        else:
            subscriptions = subscriptions if isinstance(subscriptions, list) else [subscriptions]
            for subscription in subscriptions:
                subscription = subscription if isinstance(
                    subscription, Subscription) else Subscription(self.connection, subscription,
                                                                  self.project_id)
                if subscription.delivery.mode == 'EMAIL':
                    subscription.execute()
                else:
                    msg = (f"Subscription '{subscription.name}' with ID '{subscription.id}' "
                           "could not be executed. Delivery mode '{subscription.delivery['mode']}'"
                           " is not supported.")
                    helper.exception_handler(msg, UserWarning)

    def available_bursting_attributes(self, content: Union[dict, "Content"]):
        """Get a list of available attributes for bursting feature, for a given
        content.

        Args:
            content: content dictionary or Content object
                (from subscription.content)
        """
        c_id = content['id'] if isinstance(content, dict) else content.id
        c_type = content['type'] if isinstance(content, dict) else content.type

        response = subscriptions_.bursting_attributes(self.connection, self.project_id, c_id,
                                                      c_type.upper())

        if response.ok:
            return response.json()['burstingAttributes']

    def available_recipients(self, content_id: Optional[str] = None,
                             content_type: Optional[str] = None,
                             content: Optional["Content"] = None,
                             delivery_type='EMAIL') -> List[dict]:
        """List available recipients for a subscription contents.
        Specify either both `content_id` and `content_type` or just `content`
        object.

        Args:
            content_id: ID of the content
            content_type: type of the content
            content: Content object
            delivery_type: The delivery of the subscription, available values
                are: [EMAIL, FILE, PRINTER, HISTORY_LIST, CACHE, MOBILE, FTP].
        """
        if content_id and content_type:
            pass
        elif isinstance(content, Content):
            content_id = content.id
            content_type = content.type
        else:
            helper.exception_handler('Specify either a content ID and type or content object.',
                                     ValueError)

        body = {
            "contents": [{
                "id": content_id,
                "type": content_type
            }],
        }

        response = subscriptions_.available_recipients(self.connection, self.project_id, body,
                                                       delivery_type)

        if response.ok and config.verbose:
            return response.json()['recipients']



from abc import ABCMeta, abstractmethod
import json
from pprint import pprint
from sys import version_info
from typing import Any, Dict, List, Optional, Union
import warnings

from pandas import DataFrame, Series

from mstrio import config
import mstrio.utils.helper as helper

from .setting_types import DeprecatedSetting, SettingValue, SettingValueFactory
from .settings_helper import convert_settings_to_byte, convert_settings_to_mega_byte
from .settings_io import CSVSettingsIO, JSONSettingsIO, PickleSettingsIO, SettingsSerializerFactory


class BaseSettings(metaclass=ABCMeta):
    """Base class for fetching and updating Project or Server Settings. The
    settings object can be optionally initialized with `connection` and `id`,
    which will automatically fetch the current settings for the specified
    object. If not specified, settings can be loaded from file using
    `import_from()` method. Settings attributes can be modified manually.
    Lastly, the settings object can be validated and applied to a given object
    (Project, Server) using the `update()` method.

    Attributes:
        settings: settings
    """
    _CONFIG: Dict = {}
    # !NOTE define in child classes
    _TYPE: Union[str, None] = None
    _READ_ONLY: List[str] = []
    _CONVERSION_MAP: Dict[str, str] = {}

    def alter(self, **settings):
        """Alter settings in a command manager like way."""
        for setting, value in settings.items():
            setattr(self, setting, value)

    def fetch(self) -> None:
        """Fetch current settings from I-Server and update this Settings
        object."""
        prepared_settings = self._fetch()
        for key, value in prepared_settings.items():
            if hasattr(self, key):
                setting_obj = getattr(self, key)
                setting_obj.value = value
                super(BaseSettings, self).__setattr__(key, setting_obj)

    @abstractmethod
    def _fetch(self):
        pass

    @abstractmethod
    def update(self) -> None:
        """Update the current settings on I-Server using this Settings
        object."""
        pass

    def _get_config(self):
        """Fetch the settings config to be used in this Settings class."""
        for setting, cfg in self._CONFIG.items():
            setting_type = cfg.get('type')
            if setting_type is None:
                self._READ_ONLY.append(setting)
            cfg.update({'name': setting})

    def compare_with_files(self, files: List[str], show_diff_only: bool = False) -> DataFrame:
        """Compare the current project settings to settings in file/files
        Args:
            files (str): List of paths to the settings files. Supported settings
                files are JSON, CSV, Pickle. Ex: "/file.csv"
            show_diff_only(bool, optional): Whether to display all settings or
                only different from first project in list.
        Returns:
            Dataframe with values of current settings and settings from files.
        """
        current = DataFrame.from_dict(self.list_properties(show_names=True), orient='index',
                                      columns=['value']).reset_index()
        base = 'current value'
        current.columns = ['setting', base]
        for proj in files:
            self.import_from(proj)
            current[proj] = self.to_dataframe()['value']
        if show_diff_only:
            index = Series([True] * len(current['setting']))
            for proj_name in files:
                compare = current[base].eq(current[proj_name])
                index = compare & index
            current = current[~index]
            if current.empty and config.verbose:
                print("There is no difference between current settings and settings from files")
        return current

    def to_csv(self, name: str) -> None:
        """Export the current project settings to the csv file.

        Args:
            name (str): Name of file
        """
        CSVSettingsIO.to_file(file=name, settings_obj=self)

    def to_json(self, name: str) -> None:
        """Export the current project settings to the json file
        Args:
            name (str): Name of file
        """
        JSONSettingsIO.to_file(file=name, settings_obj=self)

    def to_pickle(self, name: str) -> None:
        """Export the current project settings to the pickle file
        Args:
            name (str): Name of file
        """
        PickleSettingsIO.to_file(file=name, settings_obj=self)

    def import_from(self, file: str) -> None:
        """Import project settings from a 'csv', 'json' or 'pickle' file.

        Args:
            file (str): Path to the file with supported extension type name.
                Ex: "<path>/file.csv"
        """
        # Extract settings from supported file
        serializer = SettingsSerializerFactory().get_serializer(file)
        settings_dict = serializer.from_file(file, self)

        # Try to validate settings by fetching settings from I-Server
        self._validate_settings(settings_dict)

        # If no Exception was raised, assign the settings to the object
        for (setting, value) in settings_dict.items():
            setattr(self, setting, value)
        if config.verbose:
            print("Settings imported from '{}'".format(file))

    def list_properties(self, show_names: bool = True) -> dict:
        """Return settings and their values as dictionary.

        Args:
            show_names: if True, return meaningful setting values, else, return
                exact setting values
        """
        if show_names:
            return {
                key: self.__dict__[key]._get_value()
                for key in sorted(self.__dict__)
                if not key.startswith('_')
            }
        else:
            return {
                key: self.__dict__[key].value
                for key in sorted(self.__dict__)
                if not key.startswith('_')
            }

    def to_dataframe(self) -> DataFrame:
        """Return a `DataFrame` object containing settings and their values."""

        df = DataFrame.from_dict(self.list_properties(), orient='index', columns=['value'])
        df.reset_index(inplace=True)
        df.rename({'index': 'setting'}, axis=1, inplace=True)
        return df

    @property
    def info(self) -> None:
        if version_info.major >= 3 and version_info.minor >= 7:
            pprint(self.list_properties(), sort_dicts=False)
        else:
            pprint(self.list_properties())

    @property
    def setting_types(self) -> DataFrame:
        df = DataFrame.from_dict(self._CONFIG, orient='index')
        df.drop(['name', 'read_only', 'reboot_rule', 'deprecated'], axis='columns', inplace=True,
                errors='ignore')
        return df

    def _validate_settings(self, settings: Optional[dict] = None, bad_setting=Warning,
                           bad_type=Warning, bulk_error=True) -> None:
        """Validate setting-value pairs and raise AttributeError or TypeError
        if invalid. If `bad_setting` or `bad_type` is of type Exception, then
        Exception is raised as soon as the first invalid pair is found. If they
        are of type Warning the validation will continue and not raise error.

        Raises:
            ValueError if `bulk_error` True, tries to evaluate all settings.
        """
        settings = settings if settings else self.list_properties(show_names=False)
        bad_settings_keys = []

        for setting, value in settings.items():
            if setting not in self._CONFIG.keys():
                msg = "Setting '{}' is not supported.".format(setting)
                helper.exception_handler(msg, bad_setting)
                bad_settings_keys.append((setting, value))
            else:
                setting_obj = getattr(self, setting)
                if isinstance(setting_obj, DeprecatedSetting):
                    continue
                else:
                    valid = setting_obj._validate_value(value, exception=not bulk_error)
                    if not valid:
                        bad_settings_keys.append((setting, value))
        if bulk_error and bad_settings_keys:
            helper.exception_handler(
                "Invalid settings: {}".format(
                    [item[0] + ': ' + str(item[1]) for item in bad_settings_keys]),
                exception_type=ValueError)

    def _prepare_settings_push(self) -> dict:

        def to_rest_format(settings: dict) -> dict:
            return {k: {'value': v} for k, v in settings.items() if k not in self._READ_ONLY}

        settings = self.list_properties(show_names=False)
        settings = convert_settings_to_byte(settings, self._CONVERSION_MAP)
        settings = to_rest_format(settings)
        return settings

    def _prepare_settings_fetch(self, settings: dict) -> dict:

        def from_rest_format(settings: dict) -> dict:
            return {k: v['value'] for k, v in settings.items()}

        settings = from_rest_format(settings)
        settings = convert_settings_to_mega_byte(settings, self._CONVERSION_MAP)
        return settings

    def _cast_type_from_obj(self, setting: str, value: Any) -> Any:
        setting_obj = getattr(self, setting, None)
        processed_value = None

        if setting_obj and not isinstance(setting_obj, DeprecatedSetting):
            try:
                processed_value = setting_obj.type(value)
            except ValueError:
                processed_value = value
        else:
            processed_value = value
        return processed_value

    def _configure_settings(self):
        """Sets up the settings object to allow for verification of
        settings."""

        self._get_config()
        for setting, cfg in self._CONFIG.items():
            factory = SettingValueFactory()
            value = factory.get_setting(cfg)
            self.__override_settings_config(value)
            super(BaseSettings, self).__setattr__(setting, value)

    def __override_settings_config(self, value: SettingValue) -> None:
        pass

    def __str__(self):
        """String interpretation of the Setting object."""
        return json.dumps((self.list_properties()), indent=4)

    def __setattr__(self, name, value):
        name = super().__getattribute__(name)
        if isinstance(name, (SettingValue, DeprecatedSetting)):
            if name.value != value or not isinstance(name.value, type(value)):
                name._validate_value(value)
                if getattr(name, 'reboot_rule', None) is not None:
                    msg = f"{name.name}: {name.reboot_rule.get('description')}"
                    warnings.warn(msg, Warning)
                name.value = value
        else:
            super(BaseSettings, self).__setattr__(name, value)



from mstrio.utils.helper import deprecation_warning

from .base_subscription import *  # NOQA
from .email_subscription import *  # NOQA

deprecation_warning(
    "importing EmailSubscription from `mstrio.distribution_services.subscription.subscription`",
    "`mstrio.distribution_services.subscription`", "11.3.4.101")  # NOSONAR



# flake8: noqa
from .search_enums import CertifiedStatus, SearchDomain, SearchPattern, SearchResultsFormat
from .object import Object, list_objects
from .predefined_folders import PredefinedFolders
from .folder import (list_folders, get_my_personal_objects_contents,
                     get_predefined_folder_contents, Folder)
from .search_operations import (CertifiedStatus, get_search_results, full_search, quick_search,
                                quick_search_from_object, SearchDomain, SearchObject,
                                SearchPattern, SearchResultsFormat, start_full_search,
                                get_search_suggestions)
from .object import Object, list_objects
from .shortcut import get_shortcuts, Shortcut, ShortcutInfoFlags



"""This is the demo script to show how administrator can manage folders, objects
and their dependencies.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import Connection
from mstrio.object_management import (Folder, list_folders, get_predefined_folder_contents,
                                      get_my_personal_objects_contents, PredefinedFolders, Object,
                                      list_objects, SearchObject, get_search_suggestions,
                                      quick_search, quick_search_from_object, full_search,
                                      start_full_search, get_search_results, SearchPattern,
                                      SearchResultsFormat)
from mstrio.project_objects.dossier import Dossier
from mstrio.utils.entity import Rights
from mstrio.types import ObjectSubTypes, ObjectTypes

base_url = "https://<>/MicroStrategyLibrary/api"
username = "some_username"
password = "some_password"
conn = Connection(base_url, username, password, project_name="MicroStrategy Tutorial",
                  login_mode=1)

# list folders from a particular project
list_folders(conn, project_id="project_id")

# list configuration-level folders
list_folders(conn)

# get contents of My Personal Objects in a specific folder
get_my_personal_objects_contents(conn, project_id="project_id")

# get contents of a pre-defined folder in a specific folder
get_predefined_folder_contents(conn, folder_type=PredefinedFolders.PUBLIC_REPORTS,
                               project_id="project_id")

# create new folder in a particular folder
new_folder = Folder.create(conn, name="New Folder", parent="parent_folder_id",
                           description="Description of New Folder")

# get folder with a given id
folder = Folder(conn, id="folder_id")

# get contents of a folder (optionally as a dictionary)
contents_objs = folder.get_contents()
contents_dict = folder.get_contents(to_dictionary=True)

# alter name and description of a folder
folder.alter(name="New name of folder", description="Folder with a new name")

# add ACL right to the folder for the user and propagate to children of folder
folder.acl_add(Rights.EXECUTE, trustees="user_id", denied=False, inheritable=True,
               propagate_to_children=True)

# remove ACL from the folder for the user
folder.acl_remove(Rights.EXECUTE, trustees="user_id", denied=False, inheritable=True)

# Delete folder. When argument `force` is set to `False` (default value), then
# deletion must be approved.
folder.delete(force=True)

# list objects of a given type
reports = list_objects(conn, ObjectTypes.REPORT_DEFINITION)
documents = list_objects(conn, ObjectTypes.DOCUMENT_DEFINITION)

# initialize an object of a given type (in this case `FOLDER`) with a given id
object = Object(conn, ObjectTypes.FOLDER, "object_id")

# alter name and description of an object
object.alter(name="New name", description="New description")

# certify object
Object(conn, ObjectTypes.REPORT_DEFINITION, "object_id").certify()

# decerttify object
Object(conn, ObjectTypes.REPORT_DEFINITION, "object_id").decertify()

# get object properties as a dictionary
object.to_dict()

# Delete objects of a given types (in this case `REPORT` and 'DOCUMENT)
# and given ids. When argument `force` is set to `False` (default value), then
# deletion must be approved.
Object(conn, ObjectTypes.REPORT_DEFINITION, "report id").delete()
Object(conn, ObjectTypes.DOCUMENT_DEFINITION, "document id").delete(force=True)

# initialize SearchObject and synchronize with server
search_object = SearchObject(conn, id="search_object_id")

# get search suggestions with the pattern set performed in all unique projects
# accross the cluster (ti takes effect only in I-Server with cluster nodes)
suggestions = get_search_suggestions(conn, project="project_id", key="search pattern",
                                     max_results=20, cross_cluster=True)

# use the stored results of the Quick Search engine to return search results
# and display them as a list (in this particular case all reports which name
# begins with 'A')
objects = quick_search(conn, "project_id", name='A', pattern=SearchPattern.BEGIN_WITH,
                       object_types=[ObjectTypes.REPORT_DEFINITION])

# perform quick search based on a predefined Search Object (include ancestors
# and acl of returned objects)
objects = quick_search_from_object(conn, "project_id", search_object, include_ancestors=True,
                                   include_acl=True)

# search the metadata for objects in a specific project that match specific
# search criteria (e.g. super cubes which name ends with `cube`) and save the
# results in I-Server memory
start_dict = start_full_search(conn, "project_id", name="cube", pattern=SearchPattern.END_WITH,
                               object_types=ObjectSubTypes.SUPER_CUBE)

# Retrieve the results of a full metadata search in a tree format which were
# previously obtained by `start_full_search`
results = get_search_results(conn, search_id=start_dict['id'],
                             results_format=SearchResultsFormat.TREE)

# perform full search in one call (instead of the two steps presented above).
# Get all documents which name contains `document` and return in a list format.
# Perform search only in the root folder and its subfolders.
results = full_search(conn, "project_id", name="document", pattern=SearchPattern.CONTAINS,
                      object_types=ObjectTypes.DOCUMENT_DEFINITION, root="folder_id")

# return cubes that are used by the given dossier (it can be performed with the
# function `full_search` or method `get_connected_cubes` from `Document` class
# or method `get_dependencies` from `Entity` class)
сubes = Dossier(conn, id="dossier_id").get_connected_cubes()
cubes = Dossier(conn, id="dossier_id").list_dependencies(
    project="project_id", object_types=[ObjectSubTypes.OLAP_CUBE, ObjectSubTypes.SUPER_CUBE])
cubes = full_search(conn, project="project_id",
                    object_types=[ObjectSubTypes.OLAP_CUBE,
                                  ObjectSubTypes.SUPER_CUBE], used_by_object_id="dossier_id",
                    used_by_object_type=ObjectTypes.DOCUMENT_DEFINITION)

# we can also list cubes which uses given dossier
cubes_using_dossier = Dossier(conn, id="dossier_id").list_dependents(
    project="project_id", object_types=[ObjectSubTypes.OLAP_CUBE, ObjectSubTypes.SUPER_CUBE])

# get all objects which uses (also recursive) metric with given id
objects = full_search(conn, project="project_id", uses_object_id="metric_id",
                      uses_object_type=ObjectTypes.METRIC, uses_recursive=True)



"""This is the demo script to show how to manage intelligent (OLAP) cubes.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import Connection
from mstrio.project_objects import OlapCube, list_olap_cubes

# get connection to an environment
base_url = "https://<>/MicroStrategyLibrary/api"
username = "some_username"
password = "some_password"
connection = Connection(base_url, username, password, project_name="MicroStrategy Tutorial",
                        login_mode=1)

# list all OLAP Cubes
olap_cubes_ = list_olap_cubes(connection)

# get OLAP Cube by its identifier
olap_cube_ = OlapCube(connection, "<cube_id>")

# list available attributes, metrics and attribute forms which can be used for
# creation of a new OLAP Cube
OlapCube.available_metrics(connection)
OlapCube.available_attributes(connection)

# create new OLAP Cube
attributes = [{
    'id': '8D679D3C11D3E4981000E787EC6DE8A4',
    'name': 'Customer',
    'type': 'attribute',
    'forms': [{
        'id': '8D67A35B11D3E4981000E787EC6DE8A4',
        'name': 'Customer DESC 1',
        'type': 'form'
    }, {
        'id': '8D67A35F11D3E4981000E787EC6DE8A4',
        'name': 'Customer DESC 2',
        'type': 'form'
    }]
}]
metrics = [{'id': '5E2B495C11D66F4BC0006CBEB630224F', 'name': 'Average Sales', 'type': 'metric'}]
new_olap_cube = OlapCube.create(connection, name="New OLAP Cube", description="Some description",
                                folder_id="<folder_id>", attributes=attributes, metrics=metrics)

# Update attributes and metrics of a newly created OLAP Cube. When cube is not
# published yet, then it is possible to add/remove metrics or attributes.
# When cube is published it is possible to only change order of attributes or
# metrics which have already been added to cube.
attributes.append({
    'id': '8D679D3511D3E4981000E787EC6DE8A4',
    'name': 'Call Center',
    'type': 'attribute'
})
metrics.append({
    'id': '785636E5442BBFE1959BE3A708284498',
    'name': 'Count of Samples',
    'type': 'metric'
})
new_olap_cube.update(attributes=attributes, metrics=metrics)

# publish newly created and updated OLAP Cube
new_olap_cube.publish()

# refresh and show status of an OLAP Cube
new_olap_cube.refresh_status()
new_olap_cube.show_status()

# export sql view of an OLAP Cube
sql_view = new_olap_cube.export_sql_view()

# unpublish an OLAP Cube
new_olap_cube.unpublish()

# Delete an OLAP Cube. When `force` argument is set to `False` (default value)
# then deletion must be approved.
new_olap_cube.delete(force=True)
